import numpy as np
from tejaas.utils.logs import MyLogger

logger = MyLogger(__name__)

def load(qnull_file):
    qnull = list()
    with open(qnull_file, 'r') as mfile:
        for line in mfile:
            l = line.strip().split()
            q = float(l[0].strip())
            qnull.append(q)
    qnull = np.array(qnull)
    qmod = qnull[np.isfinite(qnull)]
    logger.debug("Read {:d} null Q-scores".format(qmod.shape[0]))
    return qmod



import os
import numpy as np
from scipy import special
from scipy import stats
import ctypes
from statsmodels.distributions.empirical_distribution import ECDF

from tejaas.utils import mpihelper
from tejaas.utils.logs import MyLogger
from tejaas.utils import project

class JPA:

    def __init__(self, x, y, comm, rank, ncore, qcalc, masks, get_pvals = False, qnull_file = None, statmodel = 'zstat', target_fdr = None):
        self.gt = x
        self.gx = y
        self._pvals = None
        self._qscores = None
        self._jpa_pvals = None
        self.rank = rank
        self.comm = comm
        self.ncore = ncore
        self.mpi = False
        if self.ncore > 1:
            self.mpi = True

        self.qcalc = qcalc
        self.masks = masks
        self.usemask = True if masks is not None else False
        self.get_empirical_pvals = False
        if get_pvals:
            self.get_empirical_pvals = True
        self.qnull_file = qnull_file
        self.statmodel = statmodel
        self.logger = MyLogger(__name__)
        self.target_fdr = target_fdr
        if self.target_fdr is not None:
            self.adj_pvals = list()
            self.pass_fdr  = list()


    @property
    def jpa_pvals(self):
        return self._jpa_pvals

    @property
    def pvals(self):
        return self._pvals

    @property
    def scores(self):
        return self._qscores


    def masked_jpascore(self, pvals, mask):
        usedgenes = np.ones_like(pvals, dtype=np.bool)
        if mask.shape[0] != 0: usedgenes[mask] = False
        res = self.jpascore(pvals[usedgenes])
        return res


    def jpascore(self, pvals):
        p = np.sort(pvals)
        n = p.shape[0]
        kmax = min(100, n)
        krange = [i + 1 for i in range(kmax)]
        digamma_n1 = special.digamma(n + 1)
        z = - ( np.log(p[:kmax]) - (special.digamma(krange) - digamma_n1) )
        zsum = np.cumsum(z)
        res = np.max(zsum)
        return res


    def clinreg_fstat(self, geno, expr, nrow):
        clib = project.get_clib('linear_regression')
        cfstat = clib.fit
        cfstat.restype = ctypes.c_int
        cfstat.argtypes = [np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                           np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                           ctypes.c_int,
                           ctypes.c_int,
                           ctypes.c_int,
                           np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED')
                          ]
    
        x = geno.reshape(-1,)
        y = expr.reshape(-1,)
        xsize = x.shape[0]
        nsnps = geno.shape[0]
        nsample = geno.shape[1]
        ngene = expr.shape[0]
        fstat = np.zeros(nsnps * ngene)
        success = cfstat(x, y, nsnps, ngene, nsample, fstat)
        res = 1 - stats.f.cdf(fstat, 1, nsample-2)
        return res


    def clinreg_zstat(self, geno, expr, nrow):
        clib = project.get_clib('linear_regression_zstat')
        czstat = clib.fit
        czstat.restype = ctypes.c_int
        czstat.argtypes = [np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                           np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                           ctypes.c_int,
                           ctypes.c_int,
                           ctypes.c_int,
                           np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED')
                          ]
    
        x = geno.reshape(-1,)
        y = expr.reshape(-1,)
        xsize = x.shape[0]
        nsnps = geno.shape[0]
        nsample = geno.shape[1]
        ngene = expr.shape[0]
        zstat = np.zeros(nsnps * ngene)
        success = czstat(x, y, nsnps, ngene, nsample, zstat)
        res = 2.0 * (1 - stats.norm.cdf(np.abs(zstat)))
        return res 


    def get_qecdf_fit(self, q, ntop):
        qecdf = ECDF(q)
        qsort = q[np.argsort(q)]
        qneg = qsort[-ntop:]
        qcut = qneg[0]
        cumsum = 0
        for qnull in qneg:
            cumsum += qnull - qcut
        lam = (1 / ntop) * cumsum
        prefact = ntop / q.shape[0]
        return qecdf, qcut, lam, prefact


    def p_qscore(self, q, qecdf, qcut, lam, prefact):
        if q < qcut:
            res = 1 - qecdf(q)
        else:
            res = prefact * np.exp(- (q - qcut) / lam)
        return res


    def get_qnull(self):
        ''' 
        This function reads the qnull file, if provided. Otherwise, generates uniform random number.
        Must be called from master and broadcast to the slaves.
        '''
        qmod = np.array([])
        if self.get_empirical_pvals:
            if self.qnull_file is not None and os.path.isfile(self.qnull_file):
                #qnull = self.read_qnull(self.qnull_file)
                self.logger.debug("Read null Q-scores from {:s}".format(self.qnull_file))
                qnull = list()
                with open(self.qnull_file, 'r') as instream:
                    for line in instream:
                        lsplit = line.strip().split()
                        q = float(lsplit[0].strip())
                        qnull.append(q)
                qnull = np.array(qnull)
            else:
                self.logger.debug("Creating null Q-scores from uniform p-value distribution")
                ngene = 10000
                nsnps = 50000
                qnull = np.array([self.jpascore(np.random.uniform(0, 1, size = ngene)) for i in range(nsnps)])
            qmod = qnull[np.isfinite(qnull)]
            self.logger.debug("Obtained {:d} null Q-scores".format(qmod.shape[0]))
        return qmod


    def slavejob(self, geno, expr, nmax, offset, masks, qnull):
        '''
        Outputs.
            pvals: p-values from every SNP-gene pair linear regression. Dimension I x G.
            qscores: JPA-score from the above p-values. Dimension I.
            p_jpa: p-values for the significance of JPA, calculated empirically. Dimension I.
        '''
        self.logger.debug('Rank {:d} calculating SNPs {:d} to {:d}'.format(self.rank, offset+1, nmax + offset))
        nsnps = geno.shape[0]
        ngene = expr.shape[0]

        # Simple linear regression calculating either f-statistic or Z-statistic for every SNP-gene pair
        if self.statmodel == 'fstat':
            pvals = self.clinreg_fstat(geno, expr, nmax)
        elif self.statmodel == 'zstat':
            pvals = self.clinreg_zstat(geno, expr, nmax)

        # Calculate JPA-score
        if self.qcalc:
            # calculate JPA for each SNP (using ngene pvals)
            if self.usemask:
                qscores = np.array([self.masked_jpascore(pvals[i*ngene : (i+1)*ngene], masks[i]) for i in range(nsnps)])
            else:
                qscores = np.array([self.jpascore(pvals[i*ngene : (i+1)*ngene]) for i in range(nsnps)])
                # nzpvals = pvals.copy()
                # zero_mask = nzpvals == 0
                # nzpmin = np.min(nzpvals[~zero_mask])
                # nzpvals[zero_mask] = nzpmin
                # qscores = np.array([self.jpascore(nzpvals[i*ngene : (i+1)*ngene]) for i in range(nsnps)])
        else:
            qscores = np.zeros(nsnps)

        # Calculate empirical p-values for the JPA-scores
        if self.get_empirical_pvals:
            ntop = min(500, int(qnull.shape[0] / 10))
            qecdf, qcut, lam, prefact = self.get_qecdf_fit(qnull, ntop)
            p_jpa = np.array([self.p_qscore(q, qecdf, qcut, lam, prefact) for q in qscores])
        else:
            p_jpa = np.array([1 for q in qscores])

        return pvals, qscores, p_jpa


    def mpicompute(self):
        if self.rank == 0:
            # this is the master
            # create a list of genotypes for sending to your slaves
            geno, offset = mpihelper.split_genotype(self.gt, self.ncore)
            expr = self.gx
            nsnp = [x.shape[0] for x in geno]
            gmasks = mpihelper.split_genemasks(self.masks, nsnp, offset)
            qnull = self.get_qnull()
        else:
            geno = None
            expr = None
            nsnp = None
            offset = None
            gmasks = None
            qnull  = None
            slave_geno = None
            slave_expr = None
            slave_nsnp = None
            slave_offs = None
            slave_gmasks = None
            slave_qnull  = None
        
        slave_geno = self.comm.scatter(geno, root = 0)
        slave_expr = self.comm.bcast(expr, root = 0)
        slave_nsnp = self.comm.scatter(nsnp, root = 0)
        slave_offs = self.comm.scatter(offset, root = 0)
        if self.usemask:
            slave_gmasks = self.comm.scatter(gmasks, root = 0)
        else:
            slave_gmasks = self.comm.bcast(gmasks, root = 0)
        slave_qnull = self.comm.bcast(qnull, root = 0)
        self.comm.barrier()
        
        # ==================================
        # Data sent. Do the calculations
        # ==================================
        pvals, qscores, p_jpa = self.slavejob(slave_geno, slave_expr, slave_nsnp, slave_offs, slave_gmasks, slave_qnull) 

        # ==================================
        # Collect the results
        # ==================================
        recvbuf = None
        if self.rank == 0:
            self.logger.debug("Number of SNPs sent to each slave: " + ", ".join(["{:d}".format(x) for x in nsnp]))
            ngene = self.gx.shape[0]
            flat_sizes = np.array([n * ngene for n in nsnp])
            recvbuf = np.zeros(sum(flat_sizes), dtype=np.float64)
        else:
            flat_sizes = None
        self.comm.Gatherv(sendbuf=pvals, recvbuf=(recvbuf, flat_sizes), root = 0)

        if self.rank == 0:
            self._pvals = recvbuf.reshape(sum(nsnp), ngene)

            ## include FDR correction
            if self.target_fdr is not None:
                self.get_fdr_each()

        qscores = self.comm.gather(qscores, root = 0)
        p_jpa = self.comm.gather(p_jpa, root = 0)

        if self.rank == 0:
            self._qscores = np.concatenate(qscores)
            if self.get_empirical_pvals:
                self._jpa_pvals = np.concatenate(p_jpa)
        else:
            assert qscores is None
            assert recvbuf is None
            assert p_jpa is None

        return
    
    def get_fdr_each(self):
        N_snp  = self.gt.shape[0]
        N_gene = self.gx.shape[0]
        for i in range(N_snp):
            snp_gene_pval = list()
            for j in range(N_gene):
                if self.usemask and j in self.masks[i]:
                    print(f"for snp {i} skipped gene {j}")
                    continue  # skip pair, gene is masked
                else:
                    snp_gene_pval.append((i,j,self._pvals[i,j]))
            pass_fdr, adj_pvals = self.bh_procedure( snp_gene_pval, self.target_fdr )
            self.pass_fdr  = self.pass_fdr + pass_fdr
            self.adj_pvals = self.adj_pvals + adj_pvals
        return

    def get_fdr_all(self):
        N_snp  = self.gt.shape[0]
        N_gene = self.gx.shape[0]
        snp_gene_pval = list()
        for i in range(N_snp):
            for j in range(N_gene):
                if self.usemask and j in self.masks[i]:
                    continue  # skip pair, gene is masked
                else:
                    snp_gene_pval.append((i,j,self._pvals[i,j]))
        self.pass_fdr, self.adj_pvals = self.bh_procedure( snp_gene_pval, self.target_fdr )
        return

    def bh_procedure(self, snp_gene_pval, target_fdr):
        self.logger.debug("Calculating FDR ... sorting {:d} SNP-gene pairs".format(len(snp_gene_pval)))
        sorted_pairs   = sorted(snp_gene_pval, key=lambda item: item[2])
        n_tests        = len(sorted_pairs) # NOT equivalent to ntrans-eqtls * ngenes, because ntrans is filtered
        pass_snps      = list()
        bh_index_limit = -1
        for i, snp_pval in enumerate(sorted_pairs[::-1]):
            bh_factor = ( (n_tests-i) / n_tests ) * target_fdr
            if snp_pval[2] > bh_factor:
                continue
            else:
                bh_index_limit = n_tests - i - 1
                break
        if bh_index_limit < 0:
            self.logger.debug("No significant SNP-gene pairs @ {:f} FDR for SNP".format(target_fdr))
            return [], []
        else:
            pass_fdr  = [ sorted_pairs[i] for i in range(bh_index_limit + 1)]
            adj_pvals = [ sorted_pairs[i][2] * ( n_tests / ( i + 1 ) ) for i in range(n_tests)] # equiv to report FDR
            return pass_fdr, adj_pvals[:len(pass_fdr)]

    def compute(self):
        if self.mpi:
            self.mpicompute()
        else:
            qnull = self.get_qnull()
            pvals, qscores, p_jpa = self.slavejob(self.gt, self.gx, self.gt.shape[0], 0, self.masks, qnull) 
            self._pvals = pvals.reshape(self.gt.shape[0], self.gx.shape[0])
            # include FDR correction
            if self.target_fdr is not None:
                self.get_fdr_each()
            self._qscores = qscores
            if self.get_empirical_pvals:
                self._jpa_pvals = p_jpa
        return



import numpy as np
from scipy.linalg import eigh
from scipy import special
from scipy import stats

from tejaas.qstats.zstats import ZSTATS
from tejaas.utils.logs import MyLogger
from tejaas.utils import mpihelper

class JPANULL:


    def __init__(self, x, y, comm, rank, ncore, outfile, niter=100000, seed=None):
        self.gt = x
        self.gx = y
        self._niter = niter
        self.outfile = outfile
        self.rank = rank
        self.comm = comm
        self.ncore = ncore
        self.mpi = False
        if seed is not None:
            np.random.seed(seed)
        if self.ncore > 1:
            self.mpi = True
        self.logger = MyLogger(__name__)


    def write_qscores(self):
        qmax = np.max(self._qscores)
        valid_qscores = self._qscores[np.where(self._qscores < qmax)]
        with open(self.outfile, 'w') as fout:
            for qnull in valid_qscores:
                fout.write(f"{qnull}\n")


    def jpascore(self, pvals):
        min_nonzero = np.min(pvals[np.nonzero(pvals)])
        pvals[pvals == 0] = min_nonzero
        p = np.sort(pvals)
        n = p.shape[0]
        kmax = min(100, n)
        krange = [i + 1 for i in range(kmax)]
        digamma_n1 = special.digamma(n + 1)
        z = - ( np.log(p[:kmax]) - (special.digamma(krange) - digamma_n1) )
        zsum = np.cumsum(z)
        res = np.max(zsum)
        return res


    def slavejob(self, W, Q, Zmean, n):
        self.logger.debug('Rank {:d} calculating {:d} null JPA scores'.format(self.rank, n))
        ngene = W.shape[0]
        pvals = np.zeros(n * ngene)
        for i in range(n):
            ngene = W.shape[0]
            zrand = np.random.normal(0, 1, size = ngene)
            znull = Zmean + np.einsum('ij, j, j', Q, np.sqrt(W), zrand)
            pvals[i*ngene : (i+1)*ngene] = 2.0 * (1 - stats.norm.cdf(np.abs(znull)))
        qnull = np.array([self.jpascore(pvals[i*ngene : (i+1)*ngene]) for i in range(n)])
        return pvals, qnull


    def mpicompute(self):
        if self.rank == 0:
            # this is the master
            # create a list of N for sending to your slaves
            thisW = self._W
            thisQ = self._Q
            thisZmean = self._Zmean
            start, end = mpihelper.split_n(self._niter, self.ncore)
            nlist = [x - y for x, y in zip(end, start)]
        else:
            thisW = None
            thisQ = None
            thisZmean = None
            nlist = None
            slave_W = None
            slave_Q = None
            slave_Zmean = None
            slave_n = None

        slave_W = self.comm.bcast(thisW, root = 0)
        slave_Q = self.comm.bcast(thisQ, root = 0)
        slave_Zmean = self.comm.bcast(thisZmean, root = 0)
        slave_n = self.comm.scatter(nlist, root = 0)
        self.comm.barrier()

        if self.rank == 0: self.logger.debug("Broadcast W, Q and Zmean to the slave nodes")

        # ==================================
        # Data sent. Do the calculations
        # ==================================
        pvals, qscores = self.slavejob(slave_W, slave_Q, slave_Zmean, slave_n)

        # ==================================
        # Collect the results
        # ==================================
        recvbuf = None
        if self.rank == 0:
            self.logger.debug("Number of SNPs sent to each slave: " + ", ".join(["{:d}".format(x) for x in nlist]))
            ngene = self._W.shape[0]
            flat_sizes = np.array([n * ngene for n in nlist])
            recvbuf = np.zeros(sum(flat_sizes), dtype=np.float64)
        else:
            flat_sizes = None
        self.comm.Gatherv(sendbuf=pvals, recvbuf=(recvbuf, flat_sizes), root = 0)

        if self.rank == 0:
            self._pvals = recvbuf.reshape(sum(nlist), ngene)

        qscores = self.comm.gather(qscores, root = 0)

        if self.rank == 0:
            self._qscores = np.concatenate(qscores)
        else:
            assert qscores is None
            assert recvbuf is None
        return


    def WQ_mpiwrap(self):
        '''
        Populates self._W, self._Q and self._Zmean
        Learns the null Zstats from the ZSTATS class
        and performs eigendecomposition in the master node.
        '''

        self._W = None
        self._Q = None
        self._Zmean = None

        if self.rank == 0: self.logger.debug("Computing Z-stats")
        zstats = ZSTATS(self.gt, self.gx, self.comm, self.rank, self.ncore)
        zstats.compute()
        
        if self.rank == 0:
            self.logger.debug("Computing W and Q")
            zscores = zstats.scores
            C = np.cov(zscores.T)
            # Numpy gives imaginary eigenvalues, use eigh from scipy
            # for decomposition of real symmetric matrix
            W, Q = eigh(C)
            self.logger.debug("Eigendecomposition done")
            # still some eigenvalues are negative. force them to zero if they are negligible. (!!!!!!!!!!!)
            # check if everything is ok
            #Wsparse = W.copy()
            #Wsparse[np.where(W < 0)] = 0
            #W = Wsparse
            W[np.where(W < 0)] = 0
            self.logger.debug("Forced negative eigenvalues to zero")
            #if not np.allclose(C, Q @ np.diag(W) @ Q.T):
            #    self.logger.error("Eigen vectors could not be forced to positive")
            #    exit
            #else:
            #    W = Wsparse
            #    self.logger.debug("Eigen vectors are forced to positive")
            Zmean = np.mean(zscores, axis = 0)
            self._W = W
            self._Q = Q
            self._Zmean = Zmean
        self.comm.barrier()


    def compute(self):
        self.WQ_mpiwrap()
        if self.rank == 0: self.logger.debug("Start MPI calculation")
        if self.mpi:
            self.mpicompute()
        else:
            pvals, qscores = self.slavejob(self._W, self._Q, self._Zmean, self._niter)
            self._pvals = pvals.reshape(self._niter, self._W.shape[0])
            self._qscores = qscores
        if self.rank == 0:
            self.logger.debug("Null JPA-scores calculated. Writing to file.")
            self.write_qscores()
        return



#/usr/bin/env python

''' Parse dosage data from VCF file.
    Usage:
       from tejaas.iotools.readvcf import ReadVCF
       readvcf = ReadVCF(vcf_filepath)
       snpinfo = readvcf.snpinfo
          ...
    Returns:
       a) information of all the variants read
            - chrom
            - pos
            - id
            - reference allele
            - alternate allele
            - minor allele frequency
       b) genotype matrix
       c) identity of the donors
'''

import numpy as np
import gzip
import os
import re
from tejaas.utils.containers import SnpInfo

class ReadVCF:


    _read_genotype_once = False


    def __init__(self, filepath, startsnp, endsnp, mode="DS", samplefile=None):
        self._filepath = filepath
        self._startsnp = startsnp
        self._endsnp = endsnp
        self._mode = mode
        if samplefile is not None:
            self._samplefile = samplefile
        else:
            self._samplefile = ""

    @property
    def snpinfo(self):
        self._run_once()
        return tuple(self._snpinfo)


    @property
    def dosage(self):
        self._run_once()
        return self._dosage


    @property
    def donor_ids(self):
        self._run_once()
        return self._donor_ids


    def _run_once(self):
        if self._read_genotype_once:
            return
        self._read_genotype_once = True
        self._read_samples()
        self._read_dosage()

    def _read_samples(self):
        if os.path.exists(self._samplefile):
            with open(self._samplefile, 'r') as samfile:
                sample = 0
                samplenames = list()
                next(samfile)
                next(samfile)
                for line in samfile:
                    if re.search('^#', line):
                        continue
                    sample += 1
                    samplenames.append(line.strip().split()[0])
            self._nsample = sample
            self._samplenames = samplenames

    def _read_dosage(self):
        dosage = list()
        snpinfo = list()
        linenum = 0
        with gzip.open(self._filepath, 'r') as vcf:
            for line in vcf:
                linestrip = line.decode().strip()
                if linestrip[:2] == '##': continue
                if linestrip[:6] == '#CHROM':
                    linesplit = linestrip.split("\t")
                    donor_ids = linesplit[9:]
                else:
                    if linenum >= self._startsnp and linenum < self._endsnp:
                        linesplit = linestrip.split("\t")
                        if linesplit[0].startswith("chr"):
                            chrom = int(linesplit[0][3:])
                        else:
                            chrom = int(linesplit[0])
                        pos   = int(linesplit[1])
                        varid = linesplit[2]
                        ref   = linesplit[3]
                        alt   = linesplit[4]

                        if self._mode == "DS":
                            if "DS" not in linesplit[8].split(':'):
                                self._mode = "GT"
                            else:
                                dsindx = linesplit[8].split(':').index("DS")
                                ds = [x.split(':')[dsindx] if len(x) > 1 else "." for x in linesplit[9:]]
                                gtindx = linesplit[8].split(':').index("GT")
                                for i, x in enumerate(ds):
                                    if x == ".":
                                        gt = linesplit[9+i].split(':')[gtindx]
                                        if len(gt) == 3 and gt[0] != "." and gt[2] != ".":
                                            ds[i] = float(int(gt[0]) + int(gt[2]))

                        if self._mode == "GT":
                            if "GT" not in linesplit[8].split(':'):
                                print("ERROR: no GT field in VCF file")
                                raise
                            gtindx = linesplit[8].split(':').index("GT")
                            gt = [x.split(':')[gtindx] if len(x) > 1 else "." for x in linesplit[9:]]
                            ds = [ float(int(x[0]) + int(x[2])) if len(x) == 3 and x[0] != "." and x[2] != "." else "." for x in gt ]

                        ds_notna = [float(x) for x in ds if x != "."]
                        freq = sum(ds_notna) / 2 / len(ds_notna)
                        maf = freq
                        snpdosage = [float(x) if x != '.' else 2 * freq for x in ds]
                        #if freq > 0.5:
                        #    maf = 1 - freq
                        #    ref = linesplit[4]
                        #    alt = linesplit[3]
                        #    snpdosage = [2 - x for x in snpdosage]

                        this_snp = SnpInfo(chrom      = chrom,
                                           bp_pos     = pos,
                                           varid      = varid,
                                           ref_allele = ref,
                                           alt_allele = alt,
                                           maf        = maf)

                        dosage.append(snpdosage)
                        snpinfo.append(this_snp)
                    if linenum > self._endsnp: break
                    linenum += 1

        if os.path.exists(self._samplefile):
            common_ids = [x for x in self._samplenames if x in donor_ids]
            print("GT Sample selection: {:d} samples were retained from a total of {:d} samples".format(len(common_ids), len(donor_ids)))
            ix = [donor_ids.index(x) for x in common_ids]
            self._dosage = np.array(dosage)[:,ix]
            self._snpinfo = snpinfo
            self._donor_ids = common_ids
        else:
            self._dosage = np.array(dosage)
            self._snpinfo = snpinfo
            self._donor_ids = donor_ids



import numpy as np
import pygtrie
import os
from tejaas.utils.logs import MyLogger

logger = MyLogger(__name__)

def load(snpinfo, maf_file = None):

    maf = [x.maf for x in snpinfo]

    if maf_file is not None and os.path.exists(maf_file):
        logger.debug("Updating MAF from user provided data. {:s}".format(maf_file))
        nmafx = 0
        nmafu = 0
        nallx = 0
        t = pygtrie.StringTrie()
        with open(maf_file, "r") as mfile:
            for line in mfile:
                l = line.split()
                try:
                    t[l[2]] = l[3:]
                except:
                    logger.warn("Error reading line {:s}".format(line))
                    pass
        for i, x in enumerate(snpinfo):
            try:
                info = t[x.varid]
                if info[1] == x.alt_allele:
                    nmafu += 1
                    maf[i] = float(info[2])
                elif info[1] == x.ref_allele:
                    nmafu += 1
                    maf[i] = 1 - float(info[2])
                else:
                    nallx += 1
                    logger.debug("Alleles do not match for {:s}".format(x.varid))
            except:
                nmafx += 1
                pass
        logger.debug("No MAF found for {:d} SNPs".format(nmafx))
        logger.debug("Allele mismatch for {:d} SNPs".format(nallx))
        logger.debug("MAF updated for {:d} SNPs".format(nmafu))
        del t

    maf = np.array(maf)

    return maf



#/usr/bin/env python
'''
Tejaas: Discover Trans-eQTLs!

This initiates the MPI calls for Tejaas.
It is a wrapper for all the options used in Tejaas.
Currently, the following arguments are processed:
1) args.jpa
    This is very similar to the CPMA-score of Brynedal et. al.
    Requires a file containing null JPA-scores.
    The user must provide a filepath using the command: --jpanull filepath
        - If file exists, then it uses the file.
        - If file does not exist, then it creates a new null file at the specified filepath.
    Performs JPA-score analysis and calculate p-values using the null JPA-scores.
3) args.rr
    Performs RR-score analysis and calculate p-values from an analytical null model.
    Two analytical null models are implemented:
        a) --null maf
            MAF null (old deprecated version, not discussed in the manuscript)
            Requires 1000G / population MAF file, see documentation.
        b) --null perm
            PERM null (null model obtained by permuting SNPs)
'''

import numpy as np
import time
import mpi4py
mpi4py.rc.initialize = False
mpi4py.rc.finalize = False
from mpi4py import MPI

from tejaas.utils.args import Args
from tejaas.utils.logs import MyLogger
from tejaas.utils import project
from tejaas.qstats.jpa import JPA
from tejaas.qstats.jpa_null import JPANULL
from tejaas.qstats.revreg import RevReg

from tejaas.iotools.data import Data
from tejaas.iotools.outhandler import Outhandler
from tejaas.iotools import readmaf
from tejaas.iotools import readqnull

from tejaas.tester import UnittestTester
from tejaas.tests.test_output_files import TestOutputFiles

def run_unittests(comm, args):
    tester = UnittestTester(TestOutputFiles)
    tester.execute()
    del tester
    return

def run_tejaas(comm, args):

    # ==================================
    # Start MPI calculation
    # ==================================
    rank = comm.Get_rank()
    ncore = comm.Get_size()
    if rank == 0: start_time = time.time()


    # ==================================
    # Input Processing
    # ==================================
    logger = MyLogger(__name__)

    # List of variables that are broadcast over all slave nodes
    gtcent       = None   # Centered genotype (x - mu). Note: Not scaled (divided) by standard deviation. Dimension I x N.
    gtnorm       = None   # Centered and scaled genotype (x - mu) / sigma. Dimension I x N.
    expr         = None   # Expression matrix. Dimension G x N.
    tgene_gtnorm = None   # Centered and scaled genotype without KNN correction. Dimension I x N.
    tgene_expr   = None   # Expression matrix for finding target genes after RR-score. Dimension G x N.
    masklist     = None   # List of gene indices masked for each SNP. List of length I. Each element contains a list of indices for the cis-genes of a candidate SNP.
    maskcomp     = None   # List of CisMasks. Each element is a collection of (a) gene indices to be masked, and (b) all SNP indices which require this mask.
    maf          = None   # List of MAF of each SNP as observed in the sample (or read separately from the population if file is provided). Length I.
    if rank == 0:
        logger.debug("Using {:d} cores".format(ncore))
        data = Data(args)
        data.load()
        gtcent       = data.geno_centered
        gtnorm       = data.geno_normed
        expr         = data.expression
        tgene_gtnorm = data.tgene_geno_normed 
        tgene_expr   = data.tgene_expression
        masklist     = data.cismasks_list
        maskcomp     = data.cismasks_comp
        snpinfo      = data.snpinfo  # slaves don't need this. Only needed for maf + outhandler in master node.
        geneinfo     = data.geneinfo # slaves don't need this. Only needed for outhandler in master node.
        maf          = readmaf.load(snpinfo, maf_file = args.maf_file)
        logger.debug("After prefilter: {:d} SNPs and {:d} genes in {:d} samples".format(gtcent.shape[0], expr.shape[0], gtcent.shape[1]))
    gtcent       = comm.bcast(gtcent, root = 0)
    gtnorm       = comm.bcast(gtnorm, root = 0)
    expr         = comm.bcast(expr,  root = 0)
    tgene_gtnorm = comm.bcast(tgene_gtnorm, root = 0)
    tgene_expr   = comm.bcast(tgene_expr, root = 0)
    masklist     = comm.bcast(masklist, root = 0)
    maskcomp     = comm.bcast(maskcomp, root = 0)
    maf          = comm.bcast(maf, root = 0)

    comm.barrier()
    if rank == 0: read_time = time.time()


    # ==================================
    # Calculate the null JPA-scores
    # ==================================
    if args.jpa_calc_null:
        if rank == 0: logger.debug("Calculating null values for JPA")
        jpa_null = JPANULL(gtnorm, expr, comm, rank, ncore, args.jpanull_file, niter=args.jpanull_iter, seed=args.seed)
        jpa_null.compute()
    if rank == 0: jpanull_time = time.time()


    # ==================================
    # Calculate JPA-score
    # with empirical p-values from the null file
    # ==================================
    if args.jpa:
        if rank == 0: logger.debug("Computing JPA")
        jpa = JPA(gtnorm, expr, comm, rank, ncore, args.jpa, masklist, get_pvals = True, qnull_file = args.jpanull_file)
        jpa.compute()

        ### Calculate target genes
        teqtl_id = None
        if rank == 0: teqtl_id = np.where(jpa.jpa_pvals <= args.psnpcut)[0]
        teqtl_id = comm.bcast(teqtl_id, root = 0)

        tgjpa = JPA(gtnorm[teqtl_id, :], expr, comm, rank, ncore, False, None, get_pvals = False, statmodel = 'fstat')
        tgjpa.compute()

    if rank == 0: jpa_time = time.time()


    # ==================================
    # Calculate the RR-score
    #   - with perm-null (default)
    #   - with MAF-null (old and deprecated, kept for legacy)
    #
    # Find target genes 
    #   - Use KNN corrected gene expression and SNPs to find SNP-gene association p-values
    #   - Use separate confounder-corrected gene expression to find SNP-gene association p-values
    #   - Run JPA only on the trans-eQTLs (discovered with user-provided cutoff) with following options:
    #       a) qcalc = False // JPA-score is not calculated
    #       b) masklist = None // do not mask any gene. we also want cis-associations if any.
    #       c) get_pvals = False // p-values for significance of JPA-score is not calculated
    #       d) statmodel = 'fstat' // Use F-statistic as used in MatrixEQTL. Default is Z-statistic as used in CPMA.
    #
    # TO-DO:
    #  - The theory of MAF-null is not included in the manuscript, create a document
    # ==================================
    if args.rr:
        sigbeta2 = np.repeat(args.sigmabeta ** 2, gtnorm.shape[0])
        if rank == 0: logger.debug("Sigma_beta fixed to {:g}".format(args.sigmabeta))
        if args.nullmodel == 'maf':
            rr = RevReg(gtnorm, expr, sigbeta2, comm, rank, ncore, null = args.nullmodel, maf = maf, masks = maskcomp)
        elif args.nullmodel == 'perm':
            rr = RevReg(gtcent, expr, sigbeta2, comm, rank, ncore, null = args.nullmodel, maf = maf, masks = maskcomp)
        rr.compute(get_betas = False)  # Set get_betas = True to obtain the coefficients of multiple regression (stored in rr.betas)

        teqtl_id = None
        if rank == 0: teqtl_id = np.where(rr.pvals <= args.psnpcut)[0]
        teqtl_id = comm.bcast(teqtl_id, root = 0)

        tgknn = JPA(gtnorm[teqtl_id, :], expr, comm, rank, ncore, False, None, get_pvals = False, statmodel = 'fstat')
        tgknn.compute()

        if args.usefdr:
            # add masks
            tgjpa = JPA(tgene_gtnorm[teqtl_id, :], tgene_expr, comm, rank, ncore, False, [masklist[i] for i in teqtl_id], get_pvals = False, statmodel = 'fstat', target_fdr=0.5)
            tgjpa.compute()
        else:
            tgjpa = JPA(tgene_gtnorm[teqtl_id, :], tgene_expr, comm, rank, ncore, False, None, get_pvals = False, statmodel = 'fstat')
            tgjpa.compute()

    if rank == 0: rr_time = time.time()


    # ==================================
    # Write output
    # ==================================
    if rank == 0: 
        ohandle = Outhandler(args, snpinfo, geneinfo)
        if args.jpa:
            ohandle.write_jpa_out(jpa, tgjpa, teqtl_id)
        if args.rr:
            ohandle.write_rr_out(rr, tgknn, tgjpa, teqtl_id, write_betas = False)
    if rank == 0: write_time = time.time()


    # ==================================
    # Log the run times
    # ==================================
    if rank == 0:
        logger.info("File reading time: {:g} seconds".format(read_time - start_time))
        logger.info("JPA null calculation time: {:g} seconds".format (jpanull_time - read_time))
        logger.info("JPA calculation time: {:g} seconds".format (jpa_time - jpanull_time))
        logger.info("RR calculation time: {:g} seconds".format (rr_time - jpa_time))
        logger.info("Result writing time: {:g} seconds".format(write_time - rr_time))
        logger.info("Total run time: {:g} seconds".format(time.time() - start_time))

def main():

    # ==================================
    # Start MPI calculation
    # ==================================
    MPI.Init()
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    ncore = comm.Get_size()

    # ==================================
    # Input Processing
    # ==================================
    args = Args(comm, rank)

    if args.maketest:
        run_unittests(comm, args)
    else:
        run_tejaas(comm, args)

    MPI.Finalize()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

''' This defines all containers used in this project
'''

import collections

SNPINFO_FIELDS = ['chrom', 'varid', 'bp_pos', 'ref_allele', 'alt_allele', 'maf']
class SnpInfo(collections.namedtuple('_SnpInfo', SNPINFO_FIELDS)):
    __slots__ = ()


GENEINFO_FIELDS = ['name', 'ensembl_id', 'chrom', 'start', 'end']
class GeneInfo(collections.namedtuple('_GeneInfo', GENEINFO_FIELDS)):
    __slots__ = ()


CISMASK_FIELDS = ['rmv_id', 'apply2']
class CisMask(collections.namedtuple('_CisMask', CISMASK_FIELDS)):
    __slots__ = ()

    @property
    def nsnp(self):
        return len(self.apply2)

    def __repr__(self):
        parent_string = super(CisMask, self).__repr__()
        return '{:s}, nsnp = {:d}'.format(parent_string, self.nsnp)



import os
import numpy as np
import ctypes
from tejaas.utils import mpihelper
from tejaas.utils.logs import MyLogger
from tejaas.utils import project

class ZSTATS:

    def __init__(self, x, y, comm, rank, ncore, masks = None):
        self.gt = x
        self.gx = y
        self._zstats = None
        self.rank = rank
        self.comm = comm
        self.ncore = ncore
        self.masks = masks
        self.usemask = True if masks is not None else False
        self.mpi = False
        if self.ncore > 1:
            self.mpi = True
        self.logger = MyLogger(__name__)


    @property
    def scores(self):
        return self._zstats


    def clinreg(self, geno, expr, nrow):
        clib = project.get_clib('linear_regression_zstat')
        czstat = clib.fit
        czstat.restype = ctypes.c_int
        czstat.argtypes = [np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                           np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                           ctypes.c_int,
                           ctypes.c_int,
                           ctypes.c_int,
                           np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED')
                          ]
    
        x = geno.reshape(-1,)
        y = expr.reshape(-1,)
        xsize = x.shape[0]
        nsnps = geno.shape[0]
        nsample = geno.shape[1]
        ngene = expr.shape[0]
        zstat = np.zeros(nsnps * ngene)
        success = czstat(x, y, nsnps, ngene, nsample, zstat)
        return zstat


    def slavejob(self, geno, expr, nmax, offset):
        self.logger.debug('Rank {:d} calculating SNPs {:d} to {:d}'.format(self.rank, offset+1, nmax + offset))
        nsnps = geno.shape[0]
        ngene = expr.shape[0]
        zstat = self.clinreg(geno, expr, nmax)
        return zstat


    def mpicompute(self):
        if self.rank == 0:
            # this is the master
            # create a list of genotypes for sending to your slaves
            geno, offset = mpihelper.split_genotype(self.gt, self.ncore)
            expr = self.gx
            nsnp = [x.shape[0] for x in geno]
        else:
            geno = None
            expr = None
            nsnp = None
            offset = None
            slave_geno = None
            slave_expr = None
            slave_nsnp = None
            slave_offs = None
        
        slave_geno = self.comm.scatter(geno, root = 0)
        slave_expr = self.comm.bcast(expr, root = 0)
        slave_nsnp = self.comm.scatter(nsnp, root = 0)
        slave_offs = self.comm.scatter(offset, root = 0)
        self.comm.barrier()

        # ==================================
        # Data sent. Do the calculations
        # ==================================
        zstat = self.slavejob(slave_geno, slave_expr, slave_nsnp, slave_offs)

        # ==================================
        # Collect the results
        # ==================================
        recvbuf = None
        if self.rank == 0:
            self.logger.debug("Number of SNPs sent to each slave: " + ", ".join(["{:d}".format(x) for x in nsnp]))
            ngene = self.gx.shape[0]
            flat_sizes = np.array([n * ngene for n in nsnp])
            recvbuf = np.zeros(sum(flat_sizes), dtype=np.float64)
        else:
            flat_sizes = None
        self.comm.Gatherv(sendbuf=zstat, recvbuf=(recvbuf, flat_sizes), root = 0)

        if self.rank == 0:
            self._zstats = recvbuf.reshape(sum(nsnp), ngene)
        else:
            assert recvbuf is None

        return
            

    def compute(self):
        if self.mpi:
            self.mpicompute()
        else:
            zstats = self.slavejob(self.gt, self.gx, self.gt.shape[0], 0)
            self._zstats = zstats.reshape(self.gt.shape[0], self.gx.shape[0])
        return



#!/usr/bin/env python

import logging
import sys
from tejaas.utils import project

loggers = {}

class MyLogger(logging.getLoggerClass()):

    global loggers

    def __init__(self, name, format="%(asctime)s | %(name)s | %(levelname)s | %(message)s", level=logging.DEBUG):

        formatter = logging.Formatter(format)
        # log to stdout
        handler = logging.StreamHandler(sys.stdout)
        # log to file
        # handler = logging.FileHandler('hello.log')
        handler.setFormatter(formatter)

        # Complete logging config.
        if loggers.get(name):
            self.logger = loggers.get(name)
        else:
            self.logger = logging.getLogger(name)
            self.logger.setLevel(level)
            self.logger.addHandler(handler)
            loggers[name] = self.logger


    def info(self, msg, extra=None):
        self.logger.info(msg, extra=extra)


    def error(self, msg, extra=None):
        self.logger.error(msg, extra=extra)


    def debug(self, msg, extra=None):
        self.logger.debug(msg, extra=extra)


    def warn(self, msg, extra=None):
        self.logger.warn(msg, extra=extra)



import numpy as np
import os
import time
import ctypes

from scipy import special
from scipy import stats
from scipy.optimize import minimize
#from statsmodels.distributions.empirical_distribution import ECDF

from tejaas.utils import mpihelper
from tejaas.qstats import rrstat
from tejaas.qstats import crrstat
from tejaas.utils.logs import MyLogger

class RevReg:

    def __init__(self, x, y, sigbeta2, comm, rank, ncore, null = 'perm', maf = None, masks = None):
        self.gt = x
        self.gx = y
        self.sigbeta2 = sigbeta2
        self.comm = comm
        self.rank = rank
        self.ncore = ncore
        self.null = null
        self.maf = maf
        self.mpi = False
        self.masks = masks
        self.usemask = False
        if self.masks is not None:
            self.usemask = True
        if self.ncore > 1:
            self.mpi = True
        self._pvals = None
        self._qscores = None
        self._mu = None
        self._sigma = None
        self._betas = None

        if self.null == 'perm':
            self.sigx2 = np.var(self.gt, axis = 1)
        elif self.null == 'maf':
            self.sigx2 = np.ones(self.gt.shape[0])

        self.logger = MyLogger(__name__)

    @property
    def sb2(self):
        return self.sigbeta2

    @sb2.setter
    def sb2(self, value):
        self.sigbeta2 = value

    @property
    def pvals(self):
        return self._pvals

    @property
    def scores(self):
        return self._qscores

    @property
    def null_mu(self):
        return self._mu

    @property
    def null_sigma(self):
        return self._sigma

    @property
    def betas(self):
        return self._betas

    def slavejob(self, gt, gx, sb2, sx2, maf, masks, start, end, usemask, get_betas = False):
        if usemask:
            if len(masks) == 0: return [], [], [], [], np.array([])
            startsnp = min([min(x.apply2) for x in masks])
            endsnp = max([max(x.apply2) for x in masks])
            totsnp = sum(x.nsnp for x in masks)
            self.logger.debug("Rank {:d} using {:d} masks on {:d} SNPs [{:d} to {:d}]".format(self.rank, len(masks), totsnp, startsnp, endsnp))
            stime = time.time()
            p, q, mu, sig, b = self.maskjob(gt, gx, sb2, sx2, maf, masks, get_betas)
            self.logger.debug("Rank {:d} took {:g} seconds".format(self.rank, time.time() - stime))
        else:
            self.logger.debug("Rank {:d}. Using {:d} SNPs [{:d} to {:d}]".format(self.rank, end - start, start, end-1))
            applyon = np.arange(start,end)
            p, q, mu, sig, b = self.basejob(gt, gx, sb2, sx2, maf, applyon, get_betas)
        return p, q, mu, sig, b

    def maskjob(self, gt, gx, sb2, sx2, maf, masks, get_betas):
        p = np.array([])
        q = np.array([])
        mu = np.array([])
        sig = np.array([])
        b = np.array([])
        for mask in masks:
            usegenes = np.ones(gx.shape[0], dtype=bool)
            if mask.rmv_id.shape[0] > 0: usegenes[mask.rmv_id] = False
            masked_gx = np.ascontiguousarray(gx[usegenes])
            _p, _q, _mu, _sig, _b = self.basejob(gt, masked_gx, sb2, sx2, maf, np.array(mask.apply2), get_betas)
            p = np.append(p, _p)
            q = np.append(q, _q)
            mu = np.append(mu, _mu)
            sig = np.append(sig, _sig)

            if get_betas:
                # set beta value for masked genes to zero
                betas = self.reshape_masked_betas(_b, mask, gx.shape[0])
                b = np.append(b, betas)

        return p, q, mu, sig, b

    def basejob(self, gt, gx, sb2, sx2, maf, applyon, get_betas):
        slv_gt  = np.ascontiguousarray(gt[applyon, :])
        slv_gx  = gx
        slv_sb2 = sb2[applyon]
        slv_sx2 = sx2[applyon]
        b = []
        if self.null == 'perm':
            p, q, mu, sig = crrstat.perm_null(slv_gt, slv_gx, slv_sb2, slv_sx2)
        elif self.null == 'maf':
            slv_maf = maf[applyon]
            p, q, mu, sig = crrstat.maf_null (slv_gt, slv_gx, slv_sb2, slv_sx2, slv_maf)
        if get_betas:
            b = crrstat.crrbetas(slv_gt, slv_gx, slv_sb2)
        #self.logger.debug("Reporting from node {:d}. Sigma = ".format(self.rank) + np.array2string(sig) + "\n" )
        return p, q, mu, sig, b


    def reshape_masked_betas(self, b, mask, ngenes):
        self.logger.debug("Rank {:d}: reshaping {:d} betas into ({:d},{:d}) with {:d} masked genes out of {:d}".format(self.rank, len(b), len(mask.apply2), (ngenes - len(mask.rmv_id)), len(mask.rmv_id), ngenes)) 
        _b = b.reshape(len(mask.apply2), ngenes-len(mask.rmv_id))
        paddedBeta = np.zeros( (len(mask.apply2), ngenes) )
        inv_ind = np.delete(np.arange(ngenes), mask.rmv_id)
        paddedBeta[:, inv_ind] = _b
        return paddedBeta.reshape(-1)


    def mpicompute(self, get_betas = False):
        gmasks = None
        pstart = None
        pend = None
        geno = None
        expr = None
        sb2  = None
        sx2  = None
        maf  = None
        if self.rank == 0:
            # this is the master
            # create a list of index for sending to your slaves
            if self.usemask:
                self.logger.debug("Masks on: " + ", ".join(['{:d}'.format(x.nsnp) for x in self.masks]))
                gmasks = mpihelper.split_maskcomp(self.masks, self.ncore)
            else:
                pstart, pend = mpihelper.split_n(self.gt.shape[0], self.ncore)

            self.logger.debug("Get betas set to "+str(get_betas))
            geno = self.gt
            expr = self.gx
            sb2  = self.sigbeta2
            sx2  = self.sigx2
            maf  = self.maf

        sb2  = self.comm.bcast(sb2,  root = 0)
        sx2  = self.comm.bcast(sx2,  root = 0)
        maf  = self.comm.bcast(maf,  root = 0)
        expr = self.comm.bcast(expr, root = 0)
        geno = self.comm.bcast(geno, root = 0)
        if self.usemask:
            gmasks = self.comm.scatter(gmasks, root = 0)
        else:
            pstart = self.comm.scatter(pstart, root = 0)
            pend = self.comm.scatter(pend, root = 0)
        self.comm.barrier()
        
        # ==================================
        # Data sent. Now do the calculations
        # ==================================
        pvals, qscores, mu, sigma, betas = self.slavejob(geno, expr, sb2, sx2, maf, gmasks, pstart, pend, self.usemask, get_betas = get_betas)

        pvals   = self.comm.gather(pvals,   root = 0)
        qscores = self.comm.gather(qscores, root = 0)
        mu      = self.comm.gather(mu,      root = 0)
        sigma   = self.comm.gather(sigma,   root = 0)

        if get_betas:
            recvbuf = None
            betalength = len(betas)
            self.comm.barrier()   # is it necessary?
            received_counts = self.comm.gather(betalength)
            if self.rank == 0:
                self.logger.debug("Number of coefficients from each node: {:s}".format(", ".join(['{:d}'.format(x) for x in received_counts])))
                recvbuf = np.zeros(np.sum(received_counts), dtype=np.float64)
            self.comm.Gatherv(sendbuf=betas, recvbuf=(recvbuf, received_counts), root = 0)

        if self.rank == 0:
            self._pvals   = np.concatenate(pvals)
            self._qscores = np.concatenate(qscores)
            self._mu      = np.concatenate(mu)
            self._sigma   = np.concatenate(sigma)
            if get_betas:
                self._betas   = recvbuf.reshape(self.gt.shape[0], self.gx.shape[0])
                self.logger.debug("All nodes computed a total of {:d} pvalues and {:s} betas".format(len(self._pvals), str(self._betas.shape)))
        else:
            assert qscores is None
            assert pvals   is None
            assert mu      is None
            assert sigma   is None
        return

    def compute(self, get_betas = False):
        if self.mpi:
            self.mpicompute(get_betas)
        else:
            start = 0
            end = self.gt.shape[0]
            self._pvals, self._qscores, self._mu, self._sigma, self._betas = self.slavejob(self.gt, self.gx,
                                                                              self.sigbeta2, self.sigx2, self.maf,
                                                                              self.masks, start, end, self.usemask, get_betas)
            if get_betas:
                self._betas = self._betas.reshape(self.gt.shape[0], self.gx.shape[0])
        return



#!/usr/bin/env python

''' Unfortunately there is no standard for gtf files.
    So, every version gets different function
'''

import os
import numpy as np
import gzip
from tejaas.utils.containers import GeneInfo


def gencode_v12(filepath, feature = 'gene', trim=False, biotype=['protein_coding'], include_chrom = 0, include_chroms=['{:d}'.format(x + 1) for x in range(22)]):
    annotfile = os.path.realpath(filepath)
    geneinfo = list()
    try:
        with gzip.open(annotfile, 'r') as mfile:
            for line in mfile:
                linesplit = line.decode().strip().split('\t')
                if linesplit[0][0] == '#' or linesplit[2] != feature: continue # skip header

                chrom = linesplit[0][3:]
                if include_chrom > 0:
                    include_chroms = ['{:d}'.format(include_chrom)]
                if chrom not in include_chroms: continue

                # Any particular biotype selected?
                infolist = linesplit[8].split(';')
                if len(biotype) > 0:
                    rowtype = infolist[2].strip().split(' ')[1].replace('"','')
                    if rowtype not in biotype: continue

                # TSS: gene start (0-based coordinates for BED)
                if linesplit[6] == '+':
                    start = np.int64(linesplit[3]) - 1
                    end   = np.int64(linesplit[4])
                elif linesplit[6] == '-':
                    start = np.int64(linesplit[3])  # last base of gene
                    end   = np.int64(linesplit[4]) - 1
                else:
                    raise ValueError('Strand not specified.')

                # For simulation
                if linesplit[1] == 'SIMULATION':
                    start = np.int64(linesplit[3])
                    end   = np.int64(linesplit[4])

                gene_id = infolist[0].strip().split(' ')[1].replace('"','')
                if trim:
                    gene_id = gene_id.split(".")[0]
                gene_name = infolist[4].strip().split(' ')[1].replace('"','')
                this_gene = GeneInfo(name       = gene_name,
                                     ensembl_id = gene_id,
                                     chrom      = int(chrom),
                                     start      = start,
                                     end        = end)

                geneinfo.append(this_gene)
    except IOError as err:
        raise IOError('{:s}: {:s}'.format(annotfile, err.strerror))

    return geneinfo



#!/usr/bin/env python

''' Unfortunately there is no standard for gtf files.
    So, every version gets different function
'''

import os
import numpy as np
import gzip
import re
from tejaas.utils.containers import GeneInfo

def gencode(filepath, feature = 'gene', trim=False, biotype=['protein_coding', 'lncRNA'], include_chrom = 0, include_chroms=['{:d}'.format(x + 1) for x in range(22)]):
    annotfile = os.path.realpath(filepath)
    geneinfo = list()
    lncRNA_list = ["macro_lncRNA", "non_coding", "bidirectional_promoter_lncRNA", "3prime_overlapping_ncRNA", 
                   "sense_overlapping", "processed_transcript", "sense_intronic", "antisense", "lincRNA", "miRNA"]

    if "lncRNA" in biotype:
        mybiotype = biotype + lncRNA_list
    else:
        mybiotype = biotype
    if annotfile.endswith("affy"):
        geneinfo = affy_exon_chip(annotfile, include_chrom = include_chrom, include_chroms=include_chroms)
        return geneinfo
    else:
        try:
            with gzip.open(annotfile, 'r') as mfile:
                for line in mfile:
                    linesplit = line.decode().strip().split('\t')
                    if linesplit[0][0] == '#' or linesplit[2] != feature: continue # skip header

                    chrom = linesplit[0][3:]
                    if include_chrom > 0:
                        include_chroms = ['{:d}'.format(include_chrom)]
                    if chrom not in include_chroms: continue

                    # Any particular biotype selected?
                    infolist = linesplit[8].split(';')
                    indices_set = False
                    gene_name_ix = None
                    gene_type_ix = None
                    gene_id_ix   = None
                    if not indices_set:
                        for i,e in enumerate(infolist):
                            Id    = e.strip().split(' ')[0]
                            # value = e.strip().split(' ')[1]
                            if Id == "gene_type": gene_type_ix = i
                            if Id == "gene_name": gene_name_ix = i
                            if Id == "gene_id": gene_id_ix = i
                        if (gene_type_ix is None) and (len(mybiotype) > 0):
                            raise ValueError("No 'gene_type' specified in gene annotation file")
                        if (gene_name_ix is not None) and (gene_id_ix is not None):
                            indices_set = True

                    if len(mybiotype) > 0:
                        rowtype = infolist[gene_type_ix].strip().split(' ')[1].replace('"','')
                        if rowtype not in mybiotype: continue
                    gene_name = infolist[gene_name_ix].strip().split(' ')[1].replace('"','')


                    # TSS: gene start (0-based coordinates for BED)
                    if linesplit[6] == '+':
                        start = np.int64(linesplit[3]) - 1
                        end   = np.int64(linesplit[4])
                    elif linesplit[6] == '-':
                        start = np.int64(linesplit[3])  # last base of gene
                        end   = np.int64(linesplit[4]) - 1
                    else:
                        raise ValueError('Strand not specified.')

                    # For simulation
                    if linesplit[1] == 'SIMULATION':
                        start = np.int64(linesplit[3])
                        end   = np.int64(linesplit[4])

                    gene_id = infolist[0].strip().split(' ')[1].replace('"','')
                    if trim:
                        gene_id = gene_id.split(".")[0]
                    this_gene = GeneInfo(name       = gene_name,
                                         ensembl_id = gene_id,
                                         chrom      = int(chrom),
                                         start      = start,
                                         end        = end)

                    geneinfo.append(this_gene)
        except IOError as err:
            raise IOError('{:s}: {:s}'.format(annotfile, err.strerror))

        ensembl_ids = [g.ensembl_id for g in geneinfo]
        try:
            assert( len(ensembl_ids) == len(set(ensembl_ids)) )
        except AssertionError:
            print('GTF annotation file contains non-unique gene_id identifiers')
        return geneinfo

def affy_exon_chip(filepath, include_chrom = 0, include_chroms=['{:d}'.format(x + 1) for x in range(22)]):
    geneinfo = list()
    try:
        with open(filepath, 'r') as mfile:
            next(mfile) # skip header
            for line in mfile:
                linesplit = line.strip().split('\t')
                if linesplit[0][0] == '#' : continue 

                chrom = linesplit[2][3:]
                if include_chrom > 0:
                    include_chroms = ['{:d}'.format(include_chrom)]
                if chrom not in include_chroms: continue

                rowtype = None
                gene_id = None
                gene_name = None
                infolist = [x.strip() for x in linesplit[8].split(';')]
                for info in infolist:
                    if info.startswith('gene_type'):
                        rowtype = info.split(' ')[1].replace('"','')
                    elif info.startswith('gene_id'):
                        gene_id = info.split(' ')[1].replace('"','')
                    elif info.startswith('gene_name'):
                        gene_name = info.split(' ')[1].replace('"','')

                # Any particular biotype selected?
                if 'all' not in biotype:
                    if rowtype not in biotype: continue

                # TSS: gene start (0-based coordinates for BED)
                if linesplit[3] == '+':
                    start = np.int64(linesplit[4]) - 1
                    end   = np.int64(linesplit[5])
                elif linesplit[3] == '-':
                    start = np.int64(linesplit[4])  # last base of gene
                    end   = np.int64(linesplit[5]) - 1
                else:
                    raise ValueError('Strand not specified.')

                gene_name = linesplit[7].split("//")[0].rstrip()
                transcript_cluster_id = linesplit[0]
                this_gene = GeneInfo(name       = gene_name,
                                     ensembl_id = transcript_cluster_id,
                                     chrom      = int(chrom),
                                     start      = start,
                                     end        = end)

                geneinfo.append(this_gene)
    except IOError as err:
        raise IOError('{:s}: {:s}'.format(annotfile, err.strerror))

    return geneinfo



import numpy as np
from sklearn.decomposition import PCA

def knn_correction(expr, dosage, K, f=0.1):
    pca = PCA(n_components=min(expr.shape[0], expr.shape[1]) )
    pca.fit(expr) # requires N x G
    expr_pca = pca.transform(expr)

    def gene_distance(a, b):
        return np.linalg.norm(a - b)

    nsample = expr.shape[0]
    distance_matrix = np.zeros((nsample, nsample))
    for i in range(nsample):
        for j in range(i+1, nsample):
            dist = gene_distance(expr_pca[i,:], expr_pca[j,:])
            distance_matrix[i, j] = dist
            distance_matrix[j, i] = dist

    kneighbor = K
    gx_knn = np.zeros_like(expr)
    gt_knn = np.zeros_like(dosage)
    neighbor_list = list()

    for i in range(nsample):
        neighbors = np.argsort(distance_matrix[i, :])[:kneighbor + 1][1:]
        gx_knn[i, :] = expr[i, :] - np.mean(expr[neighbors, :], axis = 0)
        # noisy_neighbors = np.random.choice(neighbors, size = int(2 * kneighbor / 3), replace = False)
        # noisy_neighbors = np.random.choice(neighbors, size = kneighbor, replace = True )
        noisy_neighbors = neighbors
        gt_knn[:, i] = dosage[:, i] - np.mean(dosage[:, noisy_neighbors], axis = 1)
        neighbor_list.append(neighbors)

    return gx_knn, gt_knn



import numpy as np
import scipy.stats as ss
import random
import os
from collections import defaultdict
from sklearn.decomposition import PCA

from tejaas.utils import cismasking
from tejaas.utils import knn

from tejaas.iotools import readgtf
from tejaas.iotools.readOxford import ReadOxford
from tejaas.iotools.readvcf import ReadVCF
from tejaas.iotools.readRPKM import ReadRPKM
from tejaas.utils.containers import GeneInfo, CisMask
from tejaas.utils.logs import MyLogger

SNP_COMPLEMENT = {'A':'T', 'C':'G', 'G':'C', 'T':'A'}

from functools import wraps
import time

def timeit(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time.time()
        result = f(*args, **kw)
        te = time.time()
        args[0].logger.debug('{:s} took: {:.6f} seconds'.format(f.__name__, te-ts))
        return result
    return wrap

class Data():

    def __init__(self, args):
        self.logger = MyLogger(__name__)
        self.args = args
        self._gtcent = None
        self._gtnorm = None
        self._snpinfo = None
        self._geneinfo = None
        self._expr = None
        self._cismaskcomp = None
        self._cismasklist = None
        self._tgene_gtnorm = None
        self._tgene_gtcent = None
        self._tgene_expr = None

    @property
    def geno_centered(self):
        return self._gtcent

    @property
    def geno_normed(self):
        return self._gtnorm

    @property
    def snpinfo(self):
        return self._snpinfo

    @property
    def geneinfo(self):
        return self._geneinfo

    @property
    def cismasks_comp(self):
        return self._cismaskcomp

    @property
    def cismasks_list(self):
        return self._cismasklist

    @property
    def expression(self):
        return self._expr

    @property
    def tgene_geno_normed(self):
        return self._tgene_gtnorm

    @property
    def tgene_geno_centered(self):
        return self._tgene_gtcent

    @property
    def tgene_expression(self):
        return self._tgene_expr


    def select_donors(self, vcf_donors, expr_donors):
        ''' Make sure that donors are in the same order for both expression and genotype
        '''
        common_donors = [x for x in vcf_donors if x in expr_donors]
        vcfmask = np.array([vcf_donors.index(x) for x in common_donors])
        exprmask = np.array([expr_donors.index(x) for x in common_donors])
        return vcfmask, exprmask


    def select_genes(self, info, names):
        ''' Select genes which would be analyzed. 
            Make sure the indices are not mixed up
        '''
        allowed = [x.ensembl_id for x in info]
        common  = [x for x in names if x in allowed]
        genes = [x for x in info if x.ensembl_id in common]
        indices = [names.index(x.ensembl_id) for x in genes]
        return genes, np.array(indices)


    def match_gx_indices(self, ref_gx, ref_donors, ref_gnames, gx, donors, gnames):
        '''Match the indices of gx with those of ref_gx
           Both gx and ref_gx are of size G x N
           G = genes (gnames), N = donors
        '''
        gidx = np.array([gnames.index(x) for x in ref_gnames if x in gnames])
        didx = np.array([donors.index(x) for x in ref_donors if x in donors])
        if (gidx.shape[0] != len(ref_gnames)) or (didx.shape[0] != len(ref_donors)):
            self.logger.error("Gene expression files have different donors and / or gene names. Please check. Program cancelled!")
            raise
        return gx[:, didx][gidx, :]

    
    def HWEcheck(self, x):
        gt = x.tolist()
        f = np.array([0] * 3)
        f[0] = gt.count(0)
        f[1] = gt.count(1)
        f[2] = gt.count(2)
        n = sum(f)
        X2 = n * ( (4 * f[0] * f[2] - f[1] ** 2) / ((2 * f[0] + f[1]) * (2 * f[2] + f[1])) )**2
        pval = 1 - ss.chi2.cdf(X2, 1)
        return pval


    def filter_snps(self, snpinfo, dosage, maf_limit = 0.01, use_hwe = False):
        # Predixcan style filtering of snps
        newsnps = list()
        newdosage = list()
        npoly = 0
        nambi = 0
        nunkn = 0
        nlowf = 0
        nlowf_actual = 0
        nhwep = 0
        nalle = 0
        for i, snp in enumerate(snpinfo):
            pos = snp.bp_pos
            refAllele = snp.ref_allele
            effectAllele = snp.alt_allele
            rsid = snp.varid
            maf = round(snp.maf, 3)
            # Actual MAF is lower / higher than population MAF because some samples have been removed
            maf_actual = sum(dosage[i]) / 2 / len(dosage[i])
            # Skip non-single letter polymorphisms
            if len(refAllele) > 1 or len(effectAllele) > 1:
                npoly += 1
                continue
            # Skip unknown alleles
            if refAllele not in SNP_COMPLEMENT or effectAllele not in SNP_COMPLEMENT:
                nalle += 1
                continue
            # Skip ambiguous strands
            if SNP_COMPLEMENT[refAllele] == effectAllele:
                nambi += 1
                continue
            # Skip unknown RSIDs
            if rsid == '.':
                nunkn += 1
                continue
            # Skip low MAF
            if not (maf >= maf_limit and maf <= (1 - maf_limit)):
                nlowf += 1
                continue
            # Skip low actual MAF
            if not (maf_actual >= maf_limit and maf_actual <= (1 - maf_limit)):
                nlowf_actual += 1
                continue
            # Check HWE
            if use_hwe:
                # Convert to integers 0, 1 or 2
                bins = [0.66, 1.33]
                intdosage = np.digitize(dosage[i], bins)
                # Remove SNPs out of HWE
                hwep = self.HWEcheck(intdosage)
                if(hwep < 0.000001):
                    nhwep += 1
                    # self.logger.debug("SNP {:s} has a HWE p-value of {:g}".format(rsid, hwep))
                    continue
            new_snp = snp._replace(maf = maf_actual)
            newsnps.append(new_snp)
            newdosage.append(dosage[i])
        self.logger.debug("Removed {:d} SNPs because of non-single letter polymorphisms".format(npoly))
        self.logger.debug("Removed {:d} SNPs because of unknown allele symbol".format(nalle))
        self.logger.debug("Removed {:d} SNPs because of ambiguous strands".format(nambi))
        self.logger.debug("Removed {:d} SNPs because of unknown RSIDs".format(nunkn))
        self.logger.debug("Removed {:d} SNPs because of low MAF < {:g}".format(nlowf, maf_limit))
        self.logger.debug("Removed {:d} SNPs because of low MAF (current)".format(nlowf_actual))
        if use_hwe: self.logger.debug("Removed {:d} SNPs because of deviation from HWE".format(nhwep))
        return newsnps, np.array(newdosage)


    def normalize_and_center_dosage(self, dosage):
        f = [snp.maf for snp in self._snpinfo]
        f = np.array(f).reshape(-1, 1)
        gtnorm = (dosage - (2 * f)) / np.sqrt(2 * f * (1 - f))
        gtcent = dosage - np.mean(dosage, axis = 1).reshape(-1, 1)
        return gtnorm, gtcent


    def load(self):
        ## Read Oxford File
        if self.args.oxf_file:
            oxf = ReadOxford(self.args.oxf_file, self.args.fam_file, self.args.startsnp, self.args.endsnp, isdosage=self.args.isdosage)
            dosage = oxf.dosage
            gt_donor_ids = oxf.samplenames
            snpinfo = oxf.snpinfo

        # Read VCF file
        if self.args.vcf_file:
            vcf = ReadVCF(self.args.vcf_file, self.args.startsnp, self.args.endsnp, samplefile=self.args.fam_file)
            dosage = vcf.dosage
            gt_donor_ids = vcf.donor_ids
            snpinfo = vcf.snpinfo

        # Read Gene Expression
        self.logger.debug("Reading expression levels for trans-eQTL discovery")
        rpkm = ReadRPKM(self.args.gx_file, self.args.gx_datafmt)
        expression = rpkm.expression
        expr_donors = rpkm.donor_ids
        gene_names = rpkm.gene_names

        # Read confounder corrected gene expression
        if self.args.gxcorr_file is not None:
            self.logger.debug("Reading expression levels for target gene discovery")
            rpkm_corr = ReadRPKM(self.args.gxcorr_file, self.args.gx_datafmt)
            exprcorr = self.match_gx_indices(expression, expr_donors, gene_names, rpkm_corr.expression, rpkm_corr.donor_ids, rpkm_corr.gene_names)

        self.logger.debug("Found {:d} genes of {:d} samples".format(expression.shape[0], expression.shape[1]))
        self.logger.debug("Reading gencode file for gene information")

        gene_info = readgtf.gencode(self.args.gtf_file, trim=self.args.gxtrim, biotype=self.args.biotype)
        if len(gene_info) == 0:
            self.logger.error("No gene annotations found in GTF file. Check feature 'gene' is present, with 'gene_id' and 'gene_name' annotations")

        # reorder donors gt and expr
        self.logger.debug("Selecting common samples of genotype and gene expression")
        self.logger.debug("Before expression selection: {:d} genes from {:d} samples".format(expression.shape[0], expression.shape[1]))
        vcfmask, exprmask = self.select_donors(gt_donor_ids, expr_donors)
        genes, indices = self.select_genes(gene_info, gene_names)       
        expression_selected = rpkm._normalize_expr(expression[:, exprmask][indices, :])
        if self.args.gxcorr_file is not None:
            exprcorr_selected = rpkm_corr._normalize_expr(exprcorr[:, exprmask][indices, :])
        self._geneinfo = genes

        dosage_masked = dosage[:, vcfmask]
        snpinfo_filtered, dosage_filtered_selected = self.filter_snps(snpinfo, dosage_masked)
        self.logger.debug("{:d} SNPs after filtering".format(len(snpinfo_filtered)))
        self._snpinfo = snpinfo_filtered

        self.logger.debug("After expression selection: {:d} genes from {:d} samples".format(expression_selected.shape[0], expression_selected.shape[1]))
        self.logger.debug("Retained {:d} samples".format(vcfmask.shape[0]))

        ### Until here, all filters have been applied and geneinfo and snpinfo reflect current data ###

        self._tgene_gtnorm, self._tgene_gtcent = self.normalize_and_center_dosage(dosage_filtered_selected)
        if self.args.gxcorr_file is not None:
            self._tgene_expr = exprcorr_selected
        else:
            self._tgene_expr = expression_selected

        if self.args.cismasking:
            self.logger.debug("Generate cis-masks for GX matrix for each SNP")
            self._cismasklist = cismasking.get_cismasklist(self._snpinfo, self._geneinfo, self.args.chrom, window=self.args.window)
            self._cismaskcomp = cismasking.compress_cismasklist(self._cismasklist)
            if self.args.crossmapfile is not None:
                self._cismaskcomp = cismasking.extend_cismask(self._geneinfo, self._cismaskcomp, self.args.crossmapfile )

        if self.args.knncorr:
            self.logger.debug("Applying KNN correction on gene expression and genotype")
            gx_corr, gt_corr = knn.knn_correction(expression_selected.T, dosage_filtered_selected, self.args.knn_nbr)
            self._expr = rpkm._normalize_expr(gx_corr.T)
            self._gtnorm, self._gtcent = self.normalize_and_center_dosage(gt_corr)
        else:
            self.logger.debug("No KNN correction.")
            self._expr = expression_selected
            self._gtnorm = self._tgene_gtnorm.copy()
            self._gtcent = self._tgene_gtcent.copy()
            # self._gtnorm, self._gtcent = self.normalize_and_center_dosage(dosage_filtered_selected)

        if self.args.shuffle:
            usedmask = [gt_donor_ids[i] for i in vcfmask]
            if self.args.shuffle_file is not None and os.path.isfile(self.args.shuffle_file):
                self.logger.warn("Shuffling genotype using supplied donor IDs")
                rand_donor_ids = [line.strip() for line in open(self.args.shuffle_file)]
            else:
                self.logger.warn("Shuffling genotype randomly")
                rand_donor_ids = usedmask.copy()
                random.shuffle(rand_donor_ids)
            rand_index = np.array([usedmask.index(x) for x in rand_donor_ids if x in usedmask])
            self._gtnorm = self._gtnorm[:, rand_index]
            self._gtcent = self._gtcent[:, rand_index]



#!/usr/bin/env python

import os
import sys
import numpy as np


def version():
    ## Get the version from version.py without importing the package
    vfn = os.path.join(os.path.dirname(__file__), '../version.py')
    exec(compile(open(vfn).read(), vfn, 'exec'))
    res = locals()['__version__']
    return res


def method_selector(method):
    onlyjpa = False
    jpa = False
    rrg = False
    if method == 'jpa':
        jpa = True
    elif method == 'rr':
        rrg = True
    ## Legacy flag, remove.
    if method == 'jpa-rr':
        rrg = True
    return jpa, rrg


def is_non_zero_file(fpath):
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0


def need_new_jpanull_file(method_is_jpa, fpath):
    return method_is_jpa and fpath is not None and not is_non_zero_file(fpath)


def find_module_path(libname):
    for p in sys.path:
        if os.path.isdir(p) and any([x.startswith(libname) for x in os.listdir(p)]):
            return p


def get_clib(cname):
    libprefix = 'libtejaas'
    libname = '{:s}_{:s}'.format(libprefix, cname)
    _path = find_module_path(libname)
    clib = np.ctypeslib.load_library(libname, _path)
    return clib



#!/usr/bin/env python


import sys
import gzip
import numpy as np
import re

from tejaas.utils.containers import SnpInfo
from tejaas.utils.logs import MyLogger

class Error(Exception):
    pass

class SAMPLE_NUMBER_ERROR(Error):
    pass

class GT_FREQS_NUMBER_ERROR(Error):
    pass

class ReadOxford:

    _read_samples_once = False
    _read_genotype_once = False
    _nloci = 0
    _nsample = 0

    # deafult is GTEx:
    #     - isdosage = True

    def __init__(self, gtfile, samplefile, startsnp=0, endsnp=1e15, isdosage=True):
        self.logger = MyLogger(__name__)
        self._gtfile = gtfile
        self._samplefile = samplefile
        self._startsnp = startsnp
        self._endsnp = endsnp
        self._isdosage = isdosage
        if self._isdosage:
            self._meta_columns = 6
        else:
            self._meta_columns = 5
        self._read_genotypes()

        
    @property
    def nsample(self):
        self._read_samples()
        return self._nsample


    @property
    def samplenames(self):
        self._read_samples()
        return self._samplenames

    
    @property
    def nloci(self):
        return self._nloci


    @property
    def snpinfo(self):
        self._read_genotypes()
        return tuple(self._snpinfo)


    @property
    def dosage(self):
        return tuple(self._dosage)


    @property
    def gtnorm(self):
        return tuple(self._gtnorm)


    @property
    def gtcent(self):
        return tuple(self._gtcent)


    def _read_samples(self):
        if self._read_samples_once:
           return
        self._read_samples_once = True
        with open(self._samplefile, 'r') as samfile:
            sample = 0
            samplenames = list()
            next(samfile)
            next(samfile)
            for line in samfile:
                if re.search('^#', line):
                    continue
                sample += 1
                samplenames.append(line.strip().split()[0])
        self._nsample = sample
        self._samplenames = samplenames


    def _read_dosages(self):
        dosage = list()
        allsnps = list()
        self.logger.info("Started reading genotype.")
        self._nloci = 0
        linenum = 0
        with gzip.open(self._gtfile, 'r') as filereader:
            for snpline in filereader:
                if linenum >= self._startsnp and linenum < self._endsnp:
                    self._nloci += 1
                    mline = snpline.split()

                    if self._isdosage:
                        ngenotypes = len(mline) - self._meta_columns
                    else:
                        ngenotypes = (len(mline) - self._meta_columns) / 3

                    if float(ngenotypes).is_integer():
                        if ngenotypes != self._nsample:
                            self.logger.error('Number of samples differ from genotypes')
                            raise SAMPLE_NUMBER_ERROR;
                    else:
                        self.logger.error('Number of columns in genotype frequencies not divisible by 3')
                        raise GT_FREQS_NUMBER_ERROR;

                    if self._isdosage:
                        snp_dosage = np.array([float(x) for x in mline[self._meta_columns:]])
                    else:
                        gt_freqs = np.array([float(x) for x in mline[self._meta_columns:]])
                        indsAA = np.arange(0,self._nsample) * 3
                        indsAB = indsAA + 1
                        indsBB = indsAB + 1
                        snp_dosage = 2*gt_freqs[indsBB] + gt_freqs[indsAB] # [AA, AB, BB] := [0, 1, 2]

                    maf = sum(snp_dosage) / 2 / len(snp_dosage)
                    try:                               ######## change to get the chrom numberfrom gtfile
                       chrom = int(mline[0])
                    except:
                       chrom = -1
                    this_snp = SnpInfo(chrom      = chrom,
                                       bp_pos     = int(mline[2]),
                                       varid      = mline[1].decode("utf-8"),
                                       ref_allele = mline[3].decode("utf-8"),
                                       alt_allele = mline[4].decode("utf-8"),
                                       maf        = maf)
                    allsnps.append(this_snp)
                    dosage.append(snp_dosage)
                linenum += 1
        return allsnps, np.array(dosage)


    def _read_genotypes(self):
        if self._read_genotype_once:
            return
        self._read_genotype_once = True
        self._read_samples() # otherwise, self._nsample is not set
        allsnps, dosage = self._read_dosages()
        self.logger.info("Found {:d} SNPs of {:d} samples.".format(self._nloci, self._nsample))
        self._dosage = dosage
        self._snpinfo = allsnps



import matplotlib
from matplotlib import cycler

## Resources from:
## https://matplotlib.org/users/customizing.html
## How to change color and plot styles? 
## https://matplotlib.org/users/dflt_style_changes.html
## matplotlib.rcParams[] = 

def banskt_presentation(black = '#333333', linewidth = 2, ticksize = 8, fontsize = 28, padding = 10, fontfamily = 'latex', colors = 'banskt'):

    if colors == 'banskt':
        mcolors = banskt_colors()
    elif colors == 'kelly':
        mcolors = kelly_colors()

    if fontfamily == 'latex':
        matplotlib.rcParams['text.latex.preamble'] = [r'\usepackage[sfdefault,scaled=.85, lining]{FiraSans}', 
                                                      r'\usepackage[cmintegrals]{newtxsf}',
                                                      r'\usepackage{microtype}',
                                                     ]
        matplotlib.rcParams['text.usetex'] = True
    elif fontfamily == 'latex-clearsans':
        matplotlib.rcParams['text.latex.preamble'] = [r'\usepackage[scaled=.86]{ClearSans}',
                                                      r'\usepackage[libertine]{newtxmath}',
                                                      r'\usepackage{microtype}',
                                                     ]
    elif fontfamily == 'system':
        matplotlib.rcParams['font.family'] = 'sans-serif'
        matplotlib.rcParams['font.sans-serif'] = 'DejaVu Sans'
        matplotlib.rcParams['mathtext.fontset'] =  'stixsans'

    # Size
    matplotlib.rcParams['figure.figsize'] = 8, 8
    
    # Fonts
    matplotlib.rcParams['font.size'] = fontsize
    matplotlib.rcParams['text.color'] = black
    matplotlib.rcParams['axes.titlesize'] = fontsize * 1.2
    
    matplotlib.rcParams['axes.labelsize'] = fontsize
    matplotlib.rcParams['axes.labelweight'] = 'normal'
    matplotlib.rcParams['axes.labelcolor'] = black
    
    matplotlib.rcParams['xtick.labelsize'] = fontsize
    matplotlib.rcParams['ytick.labelsize'] = fontsize
    matplotlib.rcParams['legend.fontsize'] = fontsize
    
    # Axes
    matplotlib.rcParams['axes.titlepad'] = 50
    matplotlib.rcParams['axes.edgecolor'] = black
    matplotlib.rcParams['axes.facecolor'] = 'white'
    matplotlib.rcParams['axes.labelpad'] = 20
    matplotlib.rcParams['axes.linewidth'] = linewidth
    
    # Legend
    matplotlib.rcParams['legend.facecolor'] = 'inherit'
    matplotlib.rcParams['legend.edgecolor'] = black
    matplotlib.rcParams['legend.frameon'] = False
    matplotlib.rcParams['legend.numpoints'] = 1
    matplotlib.rcParams['legend.scatterpoints'] = 1
    matplotlib.rcParams['legend.markerscale'] = 1.0
    # Dimensions as fraction of fontsize
    matplotlib.rcParams['legend.borderpad'] = 0
    matplotlib.rcParams['legend.labelspacing'] = 0.3
    matplotlib.rcParams['legend.handlelength'] = 0.5
    matplotlib.rcParams['legend.handleheight'] = 0.9
    matplotlib.rcParams['legend.handletextpad'] = 0.5
    
    # Ticks
    matplotlib.rcParams['xtick.major.top'] = False
    matplotlib.rcParams['xtick.major.bottom'] = True
    matplotlib.rcParams['xtick.minor.top'] = False
    matplotlib.rcParams['xtick.minor.bottom'] = False
    matplotlib.rcParams['ytick.major.left'] = True
    matplotlib.rcParams['ytick.major.right'] = False
    matplotlib.rcParams['ytick.minor.left'] = False
    matplotlib.rcParams['ytick.minor.right'] = False
    
    matplotlib.rcParams['xtick.major.size'] = ticksize
    matplotlib.rcParams['xtick.minor.size'] = 2 * ticksize / 3.0
    matplotlib.rcParams['ytick.major.size'] = ticksize
    matplotlib.rcParams['ytick.minor.size'] = 2 * ticksize / 3.0
    matplotlib.rcParams['xtick.major.pad'] = padding
    matplotlib.rcParams['xtick.minor.pad'] = padding
    matplotlib.rcParams['ytick.major.pad'] = padding
    matplotlib.rcParams['ytick.minor.pad'] = padding
    matplotlib.rcParams['xtick.major.width'] = linewidth
    matplotlib.rcParams['xtick.minor.width'] = linewidth
    matplotlib.rcParams['ytick.major.width'] = linewidth
    matplotlib.rcParams['ytick.minor.width'] = linewidth
    matplotlib.rcParams['xtick.color'] = black
    matplotlib.rcParams['ytick.color'] = black

    # Color cycle
    matplotlib.rcParams['axes.prop_cycle'] = cycler('color', mcolors)

    # Histogram
    matplotlib.rcParams['hist.bins'] = 20

    # Patches
    # matplotlib.rcParams['patch.facecolor'] = mcolors[0] # doesn't have any effect, comes from prop_cycle
    matplotlib.rcParams['patch.edgecolor'] = black
    matplotlib.rcParams['patch.linewidth'] = linewidth / 2
    matplotlib.rcParams['patch.force_edgecolor'] = True
    
    # For scatter plot, show only left and bottom axes
    matplotlib.rcParams['axes.spines.left'] = True
    matplotlib.rcParams['axes.spines.bottom'] = True
    matplotlib.rcParams['axes.spines.top'] = False
    matplotlib.rcParams['axes.spines.right'] = False

    return

def banskt_colors():
    banskt_colors_hex = [
        '#2D69C4', # blue 
        '#CC2529', # red
        '#93AA00', # Vivid Yellowish Green
        '#535154', # gray
        '#6B4C9A', # purple
        '#FFB300', # Vivid Yellow
        '#922428', # dark brown
        '#948B3D', # olive
        ]
    return banskt_colors_hex

def kelly_colors():
    kelly_colors_hex = [
        '#FFB300', # Vivid Yellow
        '#803E75', # Strong Purple
        '#FF6800', # Vivid Orange
        '#A6BDD7', # Very Light Blue
        '#C10020', # Vivid Red
        '#CEA262', # Grayish Yellow
        '#817066', # Medium Gray

        # The following don't work well for people with defective color vision
        '#007D34', # Vivid Green
        '#F6768E', # Strong Purplish Pink
        '#00538A', # Strong Blue
        '#FF7A5C', # Strong Yellowish Pink
        '#53377A', # Strong Violet
        '#FF8E00', # Vivid Orange Yellow
        '#B32851', # Strong Purplish Red
        '#F4C800', # Vivid Greenish Yellow
        '#7F180D', # Strong Reddish Brown
        '#93AA00', # Vivid Yellowish Green
        '#593315', # Deep Yellowish Brown
        '#F13A13', # Vivid Reddish Orange
        '#232C16', # Dark Olive Green
        ]
    return kelly_colors_hex



#!/usr/bin/env python

''' Read expression data.
    Every row contains gene expression of a donor for all genes.
    First 2 columns are FID and IID. 
    Gene expression from 3rd column onwards.
    Tab-separated.
    Reads annotation file for gene information
 
    Usage:
        from tejaas.iotools.readrpkm import ReadRPKM
        rpkm = ReadRPKM(rpkmfile)
        rpkm.read_annotation(annotationfile)
        expression = rpkm.expression
           ...

   Returns:
        a) expression matrix
        b) identity of the donors
        c) gene names
'''

import numpy as np
import os
import pandas as pd
from sklearn.decomposition import PCA

from tejaas.utils.logs import MyLogger

class ReadRPKM:


    _read_expression_once = False
    _read_annotation_once = False


    def __init__(self, filepath, dataset, npca = 0):
        self._rpkmfile = os.path.realpath(filepath)
        self._dataset = dataset
        self._genenames = None
        self._npca = npca
        self.logger = MyLogger(__name__)


    @property
    def expression(self):
        self._run_once()
        return self._gene_expression


    @property
    def donor_ids(self):
        self._run_once()
        return self._donor_ids


    @property
    def gene_names(self):
        self._run_once()
        return self._genenames


    def _run_once(self):
        if self._read_expression_once:
            return
        self._read_expression_once = True
        self._read_gene_expression()

    def _read_cardiogenics(self):
        gene_list = self._read_gene_names()
        expr_list = list()
        donor_list = list()
        with open(self._rpkmfile) as mfile:
            for line in mfile:
                linesplit = line.strip().split("\t")
                donor = linesplit[0].strip()
                expr = np.array([float(x) for x in linesplit[2:]])
                expr_list.append(expr)
                donor_list.append(donor)
        return expr_list, donor_list
    
    def _read_gtex(self):
        expr_list = list()
        donor_list = list()
        gene_list = list()
        with open(self._rpkmfile) as mfile:
            donor_list = mfile.readline().strip().split("\t")[1:]
            for line in mfile:
                linesplit = line.strip().split("\t")
                gene = linesplit[0].strip()
                gene_list.append(gene)
                expr = np.array([float(x) for x in linesplit[1:]])
                expr_list.append(expr)
        expr_list = np.transpose(np.array(expr_list))
        return gene_list, expr_list, donor_list

    def _read_geuvadis(self):
        gene_list = self._read_gene_names()
        expr_list = list()
        donor_list = list()
        with open(self._rpkmfile) as mfile:
            for line in mfile:
                linesplit = line.strip().split("\t")
                donor = linesplit[1].strip()
                expr = np.array([float(x) for x in linesplit[2:]])
                expr_list.append(expr)
                donor_list.append(donor)
        expr_list = np.array(expr_list)
        return expr_list, donor_list

    def _normalize_expr(self, Y):
        if isinstance(Y, pd.DataFrame):
            Y_cent = (Y.values - np.mean(Y.values, axis = 1).reshape(-1, 1)) / np.std(Y.values, axis = 1).reshape(-1, 1)
            Y_cent = pd.DataFrame(Y_cent, index=Y.index, columns=Y.columns)
            Y_cent.index.name = Y.index.name
        else:
            Y_cent = (Y - np.mean(Y, axis = 1).reshape(-1, 1)) / np.std(Y, axis = 1).reshape(-1, 1)
        return Y_cent

    def _normalize_expr_old(self, Y):
        newY = (Y - np.mean(Y, axis = 1).reshape(-1, 1)) / np.std(Y, axis = 1).reshape(-1, 1)
        return newY

    def _quant_normalize_expr(self, Y):
        from sklearn.preprocessing import normalize
        from sklearn.preprocessing import QuantileTransformer

        Y_quant = QuantileTransformer(output_distribution='normal').fit_transform(Y.T).T
        return Y_quant

    def _read_gene_expression(self):
        expr_list = list()
        donor_list = list()
        try:
            if self._dataset == "cardiogenics":
                gene_list, expr_list, donor_list = self._read_cardiogenics()
            if self._dataset == "gtex":
                gene_list, expr_list, donor_list = self._read_gtex()
            if self._dataset == "geuvadis":
                gene_list, expr_list, donor_list = self._read_geuvadis()
        except IOError as err:
            raise IOError("{:s}: {:s}".format(self._rpkmfile, err.strerror))
        expression = np.array(expr_list).transpose()
        normexpr = self._normalize_expr(expression)

        # if self._npca > 0:
        #     ## https://stats.stackexchange.com/questions/229092
        #     nComp = self._npca
        #     self.logger.debug("Using {:d} principal components".format(nComp))
        #     pca = PCA()
        #     pca.fit(expression.T)
        #     expr_pcacorr = np.dot(pca.transform(expression.T)[:, nComp:], pca.components_[nComp:,:]).T
        #     expression = expr_pcacorr
            
        self._genenames = gene_list
        self._gene_expression = normexpr
        self._donor_ids = donor_list


    def _read_gene_names(self):
        genefile = os.path.splitext(self._rpkmfile)[0] + ".geneids"
        genenames = list()
        try:
            with open(genefile) as mfile:
                for line in mfile:
                    gene = line.strip().split("\t")[0]
                    genenames.append(gene)
        except IOError as err:
            raise IOError("{:s}: {:s}".format(genefile, err.strerror))
        return genenames



import numpy as np
import ctypes
import os

from tejaas.utils import project

def crevreg(geno, expr, sb2, null, maf):
    clib = project.get_clib('reverse_regression')
    cqscore = clib.qscore
    cqscore.restype = ctypes.c_bool
    cqscore.argtypes = [np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        ctypes.c_int,
                        ctypes.c_int,
                        ctypes.c_int,
                        ctypes.c_int,
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED')
                       ]
    x = geno.reshape(-1,)
    y = expr.reshape(-1,)
    xsize = x.shape[0]
    nsnps = geno.shape[0]
    nsamples = geno.shape[1]
    ngenes = expr.shape[0]
    R = np.zeros(nsnps)
    p = np.zeros(nsnps)
    mu = np.zeros(nsnps)
    sigma = np.zeros(nsnps)
    success = cqscore(x, y, sb2, ngenes, nsnps, nsamples, null, maf, R, p, mu, sigma)
    return p, R, mu, sigma

def crrbetas(geno, expr, sb2):
    clib = project.get_clib('reverse_regression')
    cbetas = clib.betas
    cbetas.restype = ctypes.c_bool
    cbetas.argtypes = [np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED'),
                        ctypes.c_int,
                        ctypes.c_int,
                        ctypes.c_int,
                        np.ctypeslib.ndpointer(ctypes.c_double, ndim=1, flags='C_CONTIGUOUS, ALIGNED')
                       ]
    x = geno.reshape(-1,)
    y = expr.reshape(-1,)
    nsnps = geno.shape[0]
    nsamples = geno.shape[1]
    ngenes = expr.shape[0]
    B = np.zeros(ngenes * nsnps)
    success = cbetas(x, y, sb2, ngenes, nsnps, nsamples, B)
    return B

def maf_null(gt, gx, sb2, sx2, maf):
    p, R, mu, sigma = crevreg(gt, gx, sb2, 1, maf)
    return p, R, mu, sigma

def perm_null(gt, gx, sb2, sx2):
    maf = np.zeros(gt.shape[0])
    p, R, mu, sigma = crevreg(gt, gx, sb2, 2, maf)
    return p, R, mu, sigma

def no_null(gt, gx, sb2, sx2):
    maf = np.zeros(gt.shape[0])
    p, R, mu, sigma = crevreg(gt, gx, sb2, 0, maf)
    return p, R, mu, sigma



import numpy as np
import itertools
import os

from tejaas.utils.logs import MyLogger
logger = MyLogger(__name__)

import mpmath
mpmath.mp.dps = 500
def pval(x): return float(1 - 0.5 * (1 + mpmath.erf(x/mpmath.sqrt(2))))

class Outhandler:

    def __init__(self, args, snpinfo, geneinfo):
        self.snpinfo  = snpinfo
        self.geneinfo = geneinfo
        self.args = args
        dirname = os.path.dirname(os.path.realpath(self.args.outprefix))
        logger.debug('Writing result in: {:s}'.format(dirname))
        if not os.path.exists(dirname): os.makedirs(dirname)


    def write_jpa_out(self, jpa, tgjpa, snp_select_idx):
        fname = self.args.outprefix + "_jpa.txt"
        scores = jpa.scores
        pvals = jpa.jpa_pvals
        with open(fname, "w") as f:
            f.write("snpid\tjpascore\tp-value\n")
            for i, snp in enumerate(self.snpinfo):
                f.write("{:s}\t{:g}\t{:g}\n".format(snp.varid, scores[i], pvals[i]))
        
        fname = self.args.outprefix + "_gene_snp_list.txt"
        with open(fname, "w") as f:
            f.write("geneid\tsnpid\tpval\n")
            for i, sidx in enumerate(snp_select_idx):
                gene_select_idx = np.where(tgjpa.pvals[i, :] < self.args.pgenecut)[0]
                for gidx in gene_select_idx:
                    f.write( "{:s}\t{:s}\t{:g}\n".format(self.geneinfo[gidx].ensembl_id, self.snpinfo[sidx].varid, tgjpa.pvals[i, gidx]) )

    def write_rr_out(self, rr, tgknn, jpa, snp_select_idx, suffix = "", write_betas = False):
        mysnpinfo = self.snpinfo
        fname = self.args.outprefix + "_rr" + suffix + ".txt"
        with open(fname, "w") as f:
            f.write("{:s}\t{:s}\t{:s}\t{:s}\t{:s}\t{:s}\t{:s}\t{:s}\n".format('ID', 'CHR', 'Pos', 'MAF', 'Q', 'Mu', 'Sigma', 'P'))
            for i, x in enumerate(mysnpinfo):
                if rr.pvals[i] == 0:
                    rr.pvals[i] = pval( (rr.scores[i] - rr.null_mu[i]) / rr.null_sigma[i])
                f.write("{:s}\t{:d}\t{:d}\t{:g}\t{:g}\t{:g}\t{:g}\t{:g}\n".format(x.varid, x.chrom, x.bp_pos, x.maf, rr.scores[i], rr.null_mu[i], rr.null_sigma[i], rr.pvals[i]))

        if rr.betas is not None and write_betas:
            betafile = self.args.outprefix + "_betas" + suffix + ".txt"
            with open(betafile, 'w') as outstream:
                gene_names = " ".join([g.ensembl_id for g in self.geneinfo])
                outstream.write(gene_names+"\n")
                np.savetxt(outstream, rr.betas[snp_select_idx,:], fmt='%1.4e')

        fname = self.args.outprefix + "_gene_snp_list_knn" + suffix + ".txt"
        with open(fname, "w") as f:
            f.write("geneid\tsnpid\tpval\n")
            for i, sidx in enumerate(snp_select_idx):
                gene_select_idx = np.where(tgknn.pvals[i, :] < self.args.pgenecut)[0]
                for gidx in gene_select_idx:
                    f.write( "{:s}\t{:s}\t{:g}\n".format(self.geneinfo[gidx].ensembl_id, self.snpinfo[sidx].varid, tgknn.pvals[i, gidx]) )

        fname = self.args.outprefix + "_gene_snp_list" + suffix + ".txt"
        with open(fname, "w") as f:
            f.write("geneid\tsnpid\tpval\n")
            for i, sidx in enumerate(snp_select_idx):
                gene_select_idx = np.where(jpa.pvals[i, :] < self.args.pgenecut)[0]
                for gidx in gene_select_idx:
                    f.write( "{:s}\t{:s}\t{:g}\n".format(self.geneinfo[gidx].ensembl_id, self.snpinfo[sidx].varid, jpa.pvals[i, gidx]) )

        if jpa.target_fdr is not None:
            fname = self.args.outprefix + "_gene_snp_list_FDR" + suffix + ".txt"
            with open(fname, 'w') as f:
                f.write("geneid\tsnpid\tpval\tadj_pval\n")
                for i, pairs in enumerate(jpa.pass_fdr):
                    f.write( "{:s}\t{:s}\t{:g}\t{:g}\n".format(self.geneinfo[pairs[1]].ensembl_id, self.snpinfo[snp_select_idx[pairs[0]]].varid, pairs[2], jpa.adj_pvals[i]) )





import sys
import os
import subprocess
import setuptools
import copy
import textwrap
from configparser import ConfigParser

from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.install   import install
from setuptools.command.build_ext import build_ext
from setuptools                   import Extension
from distutils                    import log

'''
--------------------------
    SETTINGS
--------------------------
All settings are in configs.ini except version number.
'''
config = ConfigParser(delimiters=['='])
config.read('configs.ini')
cfg = config['metadata']
sys.path.append('src')
sys.path.append('src/{:s}'.format(cfg['name']))
from version import __version__

def cfg_string_tolist(key, sep):
    return [x.strip() for x in cfg[key].split(sep) if len(x) > 0]

'''
--------------------------
    LINEAR ALGEBRA
--------------------------
We want to get information about the linear algebra library
installed in the system.
Hack: piggyback on the numpy setuptools, 
      copy system_info.py and cpuinfo.py
It provides information about various resources (libraries, library directories,
include directories, etc.) in the system. Usage:
    info_dict = get_info(<name>)
  where <name> is a string 'atlas','x11','fftw','lapack','blas',
  'lapack_src', 'blas_src', etc. For a complete list of allowed names,
  see the definition of get_info() function in src/system_info.py
  Returned info_dict is a dictionary which is compatible with
  distutils.setup keyword arguments. If info_dict == {}, then the
  asked resource is not available (system_info could not find it).
  Several *_info classes specify an environment variable to specify
  the locations of software. When setting the corresponding environment
  variable to 'None' then the software will be ignored, even when it
  is available in system.

WARN: Numpy is not installed by default in a Python environment.

We have to make sure `numpy` is installed before the setup.
Possible ways of doing this:
  - define `numpy` in 'setup_requires' (instead of 'install_requires') of setuptools.setup()
  - `pyproject.toml`.

NOTE: I originally used get_info directly from numpy
    from numpy.distutils.system_info import get_info
However, the info is incorrect for AMD / MKL, 
hence I copied the files from numpy and used a hack to 
include Intel MKL on AMD machines.
(see line 1186 in src/system_info.py)
We can later move to numpy distribution of system_info
when the bug is fixed.

Obtain information about system Lapack.
The default order for the libraries are:
  - MKL
  - OpenBLAS
  - libFLAME
  - ATLAS
  - LAPACK (NetLIB)
'''
def numpy_get_lapack_info():
    from system_info import get_info
    return get_info('lapack_opt', 0)


'''
--------------------------
    C LIBRARIES
--------------------------
The sources and target shared libraries are defined in the config file.
Example:
  clib_path = src/tejaas/clibs
  clib_sources = 
    linear_regression       :: linear_regression.c
    linear_regression_zstat :: linear_regression_zstat.c
    reverse_regression      :: reverse_regression.c
LAPACK installation of the system is obtained from numpy.distutils.system_info (see above)
and these sources are combined to obtain a list of extension modules
'''

def cfg_clib_todict(clibs, clib_dir, libprefix):
    ''' 
    Reads cfg input format of clibs
    clibs is a list with strings, formatted as
        name :: filea.c, fileb.c, ...
    Returns dict {name: list<source_files>}
    '''
    clib_dict = dict()
    for clib in clibs:
        strsplit = clib.split('::')
        libname = '{:s}_{:s}'.format(libprefix, strsplit[0].strip())
        libsources = [x.strip() for x in strsplit[1].split(',')]
        clib_dict[libname] = [os.path.join(clib_dir, x) for x in libsources]
    return clib_dict


def compile_extension_dict (name, sources, extra_libraries, extra_compile_args, **kw):

    def dict_append(d, **kws):
        for k, v in kws.items():
            if k in d:
                ov = d[k]
                if isinstance(ov, str):
                    d[k] = v
                else:
                    d[k].extend(v)
            else:
                d[k] = v

    ext_args = copy.copy(kw)
    ext_args['name'] = name
    ext_args['sources'] = sources

    if 'extra_info' in ext_args:
        extra_info = ext_args['extra_info']
        del ext_args['extra_info']
        if isinstance(extra_info, dict):
            extra_info = [extra_info]
        for info in extra_info:
            dict_append(ext_args, **info)

    # Add extra libraries / compile_args 
    libraries = ext_args.get('libraries', [])
    ext_args['libraries'] = libraries + extra_libraries
    ext_args['extra_compile_args'] = extra_compile_args
    #
    return Extension(**ext_args)


def ext_modules():
    '''
    Compiles the ext_modules
    '''
    cwd         = os.path.abspath(os.path.dirname(__file__))
    clib_dir    = os.path.join(cwd, cfg['clib_path'])
    clib_cfgs   = cfg_string_tolist('clib_sources', '\n')
    lapack_info = numpy_get_lapack_info()
    print(lapack_info)
    # Prefix to be added to shared libraries
    libprefix   = 'lib{:s}'.format(cfg['name'])
    # 
    extra_libraries = []
    extra_compile_args = ['-O3', '-Werror=implicit-function-declaration']
    #
    clibs       = cfg_clib_todict(clib_cfgs, clib_dir, libprefix)
    modules     = [compile_extension_dict(k, v, extra_libraries, extra_compile_args,
                      **dict(extra_info = lapack_info)) \
                      for k, v in clibs.items()]
    return modules


'''
--------------------------
    CUSTOM OVERRIDES
--------------------------
Example custom overrides,
which can be useful to modify the build process.
'''
class custom_build_ext(build_ext):

    def build_extensions(self):
        # self.compiler - the system compiler identified by setuptools / distutils
        # print(vars(self.compiler))
        # for compiler_arg in vars(self.compiler):
        #     print (compiler_arg)
        #     print (getattr(self.compiler, compiler_arg))

        log.info("Building C modules")
        for ext in self.extensions:
            log.info(ext.name)
        #    print (dir(ext))
            property_list = ['define_macros', 'depends', 'export_symbols', 'extra_compile_args', 
                             'extra_link_args', 'extra_objects', 'include_dirs', 'libraries', 
                             'library_dirs', 'runtime_library_dirs', 'sources', 'swig_opts', 
                             'undef_macros']
            #log.info("Name: ", ext.name)
            #log.info("Language: " , ext.language)
            #log.info("py_limited_api: ", ext.py_limited_api)
            #for info in property_list:
            #    log.info(f"{info}: " + ", ".join(getattr(ext, info)))
        build_ext.build_extensions(self)


class bdist_egg_disabled(bdist_egg):
    '''
    Disabled version of bdist_egg
    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    '''
    def run(self):
        sys.exit("ERROR: aborting implicit building of eggs. Use \"pip install .\" to install from source.")

'''
--------------------------
    SETUP
--------------------------
'''
def setup_package():

    if cfg['release_branch'] == 'True':
        req_np = 'numpy>={},<{}'.format(cfg['min_numpy'], cfg['max_numpy'])
        req_py = '>={},<{}'.format(cfg['min_python'], cfg['max_python'])
    else:
        req_np = 'numpy>={}'.format(cfg['min_numpy'])
        req_py = '>={}'.format(cfg['min_python'])

    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    cmdclass = {
        'bdist_egg': bdist_egg if 'bdist_egg' in sys.argv else bdist_egg_disabled,
        'build_ext': custom_build_ext,
    }

    console_scripts = [x.replace('::', '=') for x in cfg_string_tolist('console_scripts', '\n')]

    print(numpy_get_lapack_info())

    #cfg_str_keys  = 'name author author_email description copyright license'.split()
    #cfg_url_keys  = 'main_url download_url git_url docs_url source_code_url bug_tracker_url'.split()
    #cfg_list_keys = 'keywords setup_requires install_requires contributiors'.split()

    metadata = dict(
        name             = cfg['name'],
        version          = __version__,
        author           = cfg['author'],
        author_email     = cfg['author_email'],
        description      = cfg['description'],
        long_description = long_description,
        long_description_content_type = "text/markdown",
        license          = cfg['license'],
        url              = cfg['main_url'],
        download_url     = cfg['download_url'],
        project_urls     = {
            "Bug Tracker":   cfg['bug_tracker_url'],
            "Documentation": cfg['docs_url'],
            "Source Code" :  cfg['source_code_url'],
        },
        classifiers      = cfg_string_tolist('classifiers', '\n'),
        packages         = setuptools.find_packages(where = "src"),
        package_dir      = {"": "src"},
        entry_points     = {'console_scripts': console_scripts},
        ext_modules      = ext_modules(),
        cmdclass         = cmdclass,
        #setup_requires   = cfg_string_tolist('setup_requires', ','),
        python_requires  = req_py,
        install_requires = [req_np] + cfg_string_tolist('install_requires', ','),
        keywords         = cfg_string_tolist('keywords', ','),
    )

    setuptools.setup(**metadata)


if __name__ == '__main__':
    setup_package()



'''
'''

import unittest

class UnittestTester:

    def __init__(self, test_class, verbosity = 1):
        self.test_class = test_class
        self.loader = unittest.TestLoader()
        if not isinstance(test_class, list):
            self.test_class = [test_class]
        else:
            self.test_class = test_class


    def _test_list(self):
        test_list = [self.loader.loadTestsFromTestCase(tclass) \
                        for tclass in self.test_class]
        return test_list


    def _test_suites(self):
        suites = unittest.TestSuite(self._test_list())
        return suites


    def _result_class(self):
        result_class = UnittestResult
        return result_class


    def execute(self):
        self.runner = unittest.TextTestRunner(resultclass = self._result_class())
        self.runner.run(self._test_suites())


'''
Example of overriding classes in the unittest.
Here, the TextTestResult is redefined to 
skip the progress dots.
(print personal log messages instead)
'''
class UnittestResult(unittest.TextTestResult):
    def addSuccess(self, test):
        super(unittest.TextTestResult, self).addSuccess(test)
        if self.showAll:
            self.stream.writeln("ok")
        elif self.dots:
            # All this for removing a dot from output :)
            #self.stream.write('.')
            self.stream.flush()


'''
Example of injecting the overriden class to the main test program.
This can be used as
    import cpydemo.unittest_tester as m_unittest
    m_unittest.main()
'''
class MTestProgram(unittest.TestProgram):
    def runTests(self):
        self.testRunner = unittest.TextTestRunner(resultclass = UnittestResult)
        super(MainTestProgram, self).runTests()

main = MTestProgram



#!/usr/bin/env python
'''
Collect and streamline all arguments and format them.
Sanity check on input values.
'''

import argparse
import os
import errno
from tejaas.utils.logs import MyLogger
from tejaas.utils import project


class InputError(Exception):
    ''' Raise when incorrect options are used in the argument '''
    pass


class LargerEndSnpError(InputError):
    ''' Raise when end SNP index is larger than start SNP index '''
    pass


def snprange(mstring):
    ''' 
    snprange: a colon-separated user input style for specifying start:end of SNPs to read from VCF file.
    This function is a parser for the colon-separated string.
    '''
    try:
        incsnps = mstring.split(":")
        startsnp  = int(incsnps[0].strip())
        endsnp    = int(incsnps[1].strip())
        mlist = [startsnp, endsnp]
        if endsnp < startsnp:
            raise LargerEndSnpError
    except IndexError:
        raise argparse.ArgumentTypeError('Value has to be a range of SNPs separated by colon')
    except LargerEndSnpError:
        raise argparse.ArgumentTypeError('End SNP position must be higher than start SNP position')
    return mlist


def biotype_fmt(stringlist):
    allowed_types = ['protein_coding', 'lncRNA']
    try:
        assert (all([x in allowed_types for x in stringlist]))
    except AssertionError:
        raise argparse.ArgumentTypeError('Please specify a correct biotype')
    return stringlist


def method_strings(mstring):
    '''
    Check if the specified method name is valid.
    '''
    try:
        assert (mstring == 'jpa' or mstring == 'rr' or mstring == 'jpa-rr')
    except AssertionError:
        raise argparse.ArgumentTypeError('Please specify a correct method')
    return mstring


def null_strings(mstring):
    '''
    Check if the specified null model is valid.
    '''
    try:
        assert (mstring == 'perm' or mstring == 'maf')
    except AssertionError:
        raise argparse.ArgumentTypeError('Please specify a correct null model')
    return mstring


class Args():

    def __init__(self, comm, rank):

        self.logger = MyLogger(__name__)
        self.rank = rank
        self.comm = comm

        args = None
        if self.rank == 0:
            args = self.parse_args()
        args = self.comm.bcast(args, root = 0)

        self.vcf_file    = args.vcf_filename
        self.oxf_file    = args.oxf_filename
        self.isdosage    = args.isdosage
        self.fam_file    = args.fam_filename
        if args.chrom is not None:
            self.chrom   = int(args.chrom)
        else:
            self.chrom   = None
        self.gx_file     = args.gx_filename
        self.gxcorr_file = args.gxcorr_filename
        self.gx_datafmt  = args.gx_datafmt
        self.gtf_file    = args.gtf_filename
        self.gxtrim      = args.gxtrim
        self.biotype     = args.biotype
        self.outprefix   = args.outprefix
        if args.incsnps is not None:
            self.startsnp = args.incsnps[0] - 1
            self.endsnp   = args.incsnps[1]
        else:
            self.startsnp  = 0
            self.endsnp    = 1e15 # an unusually high number to ensure all SNPs are read.

        self.jpa, self.rr = project.method_selector(args.method)
        self.nullmodel   = args.nullmodel
        self.cismasking  = args.cismasking
        self.window      = args.window
        self.sigmabeta   = args.sigmabeta
        self.knn_nbr     = args.knn
        self.knncorr     = True
        if args.knn == 0:
            self.knncorr = False

        self.shuffle        = args.shuffle
        self.shuffle_file   = args.shuffle_file
        if self.shuffle_file is not None:
            self.shuffle = True

        self.psnpcut   = args.psnpthres
        self.pgenecut  = args.pgenethres
        self.maf_file  = args.maf_filename
        self.jpanull_file = args.qnullfile
        self.jpa_calc_null = project.need_new_jpanull_file(self.jpa, self.jpanull_file)
        self.jpanull_iter = args.qnull_iter
        self.seed      = args.seed

        self.maketest  = args.maketest

        if not self.maketest: self.check_inputs()
        self.crossmapfile = args.crossmapfile
        self.usefdr = False
        self.target_fdr = args.target_fdr
        if self.target_fdr is not None:
            self.usefdr = True


        if self.rank == 0:
            self.logger.info('Method: {:s}'.format(args.method))
            if self.rr:
                self.logger.info('Null Model: {:s}'.format(args.nullmodel))
                self.logger.info('Sigma_beta: {:g}'.format(args.sigmabeta))


    def parse_args(self):

        self.logger.info('Running TEJAAS v{:s}'.format(project.version()))

        parser = argparse.ArgumentParser(description='Tejaas: Discover trans-eQTLs!')
    
        parser.add_argument('--vcf',
                            type=str,
                            dest='vcf_filename',
                            metavar='FILE',
                            help='Input VCF file in vcf.gz format')

        parser.add_argument('--oxf',
                            type=str,
                            dest='oxf_filename',
                            metavar='FILE',
                            help='Input Oxford file')

        parser.add_argument('--dosage',
                            dest='isdosage',
                            action='store_true',
                            help='Read dosages')

        parser.add_argument('--fam',
                            type=str,
                            dest='fam_filename',
                            metavar='FILE',
                            help='Input fam file')
    
        parser.add_argument('--chrom',
                            dest='chrom',
                            metavar='NUMBER',
                            help="Chromosome number of the genotype file")

        parser.add_argument('--include-SNPs',
                            type=snprange,
                            dest='incsnps',
                            metavar='START:END',
                            help='Colon-separated index of SNPs to be included')
    
        parser.add_argument('--gx',
                            type=str,
                            dest='gx_filename',
                            metavar='FILE',
                            help='input expression file for finding trans-eQTLs')

        parser.add_argument('--gxcorr',
                            type=str,
                            dest='gxcorr_filename',
                            metavar='FILE',
                            help='input expression file for finding target genes')

        parser.add_argument('--gxfmt',
                            type=str,
                            dest='gx_datafmt',
                            metavar='GX_FORMAT',
                            default='gtex',
                            help='Format of input gene expression file. Supported: gtex, cardiogencis and geuvadis')

        parser.add_argument('--biotype',
                            nargs='*',
                            type=biotype_fmt,
                            dest='biotype',
                            metavar='BIOTYPE_OPTIONS',
                            default=['protein_coding', 'lncRNA'],
                            help='List of biotypes to be selected from the GENCODE annotation file. Supported options: protein_coding, lncRNA')

        parser.add_argument('--gtf',
                            type=str,
                            dest='gtf_filename',
                            metavar='FILE',
                            help='input gtf file')
    
        parser.add_argument('--trim',
                            dest='gxtrim',
                            action='store_true',
                            help='Trim version number from GENCODE Ensembl IDs')

        parser.add_argument('--outprefix',
                            type=str,
                            dest='outprefix',
                            default='out',
                            metavar='STR',
                            help='prefix for all output files')

        parser.add_argument('--method',
                            default='rr',
                            type=method_strings,
                            dest='method',
                            metavar='STR',
                            help='which method to run: jpa / rr')
    
        parser.add_argument('--null',
                            default='perm',
                            type=null_strings,
                            dest='nullmodel',
                            metavar='STR',
                            help='which null model to use: perm / maf')
    
        parser.add_argument('--cismask',
                            dest='cismasking',
                            action='store_true',
                            help='Generate cismasks for the expression matrix for each SNP')

        parser.add_argument('--window',
                            type = int,
                            default = 1e6,
                            dest = 'window',
                            help = 'Window (number of base pairs) used for masking cis genes')

        parser.add_argument('--prior-sigma',
                            default=0.1,
                            type=float,
                            dest='sigmabeta',
                            metavar='FLOAT',
                            help='standard deviation of the normal prior for reverse multiple linear regression')

        parser.add_argument('--knn',
                            type = int,
                            dest = 'knn',
                            help = 'Number of neighbours for KNN (use 0 if you do not want KNN correction)',
                            default = 0)

        parser.add_argument('--psnpthres',
                            default=0.0001,
                            type=float,
                            dest='psnpthres',
                            metavar='PVAL',
                            help='target genes will be reported for trans-eQTLs, which are below this threshold p-value for RR/JPA statistics')
    
        parser.add_argument('--pgenethres',
                            default=0.001,
                            type=float,
                            dest='pgenethres',
                            metavar='PVAL',
                            help='target genes whose linear regression association with trans-eQTLs are below this threshold p-value will be reported')
    
        parser.add_argument('--jpanull',
                            type = str,
                            dest = 'qnullfile',
                            help = 'Filename for storing / reading null JPA scores')

        parser.add_argument('--jpanull-iter',
                            default = 100000,
                            type = int,
                            dest = 'qnull_iter',
                            help = 'Number of iterations for creating null JPA scores')

        parser.add_argument('--seed',
                            default = None,
                            type = int,
                            dest = 'seed',
                            help = 'Seed the random generator for numpy, used for development purpose')

        parser.add_argument('--maf-file',
                            type=str,
                            dest='maf_filename',
                            metavar='FILE',
                            help='file name of the MAF, see Documentation for filetype')
    
        parser.add_argument('--shuffle',
                            dest='shuffle',
                            action='store_true',
                            help='Shuffle the genotypes randomly')

        parser.add_argument('--shuffle-with',
                            type=str,
                            dest='shuffle_file',
                            metavar='FILE',
                            help='Shuffle the genotypes using the supplied donor IDs file')

        parser.add_argument('--test',
                            dest='maketest',
                            action='store_true',
                            help='whether to do test run')

        parser.add_argument('--crossmap',
                            type = str,
                            default=None,
                            dest = 'crossmapfile',
                            help = 'Crossmapability file (Saha, Battle 2018) ')

        parser.add_argument('--fdrgenethres',
                            default=None,
                            type=float,
                            dest='target_fdr',
                            metavar='FDR',
                            help='enable FDR correction up to a certain cutoff for target gene discovery')

        res = parser.parse_args()
        return res


    def check_inputs(self):
        '''
        Perform sanity checks on the input options.
        '''
        if self.rank == 0:
            '''
            Check if any genotype file is specified.
            '''
            try:
                assert (self.vcf_file is not None) or (self.oxf_file is not None)
            except AssertionError:
                print ('Input error: Specify either --vcf or --oxf. See --help for details.')
                raise

            if (self.oxf_file is not None):
                try:
                    assert (self.fam_file is not None)
                except AssertionError:
                    print ('Input error: Specify the sample file with --fam. See --help for details.')
                    raise

            '''
            Check if gene expression and GTF files are specified
            '''
            try:
                assert (self.gx_file is not None)
            except AssertionError:
                print ('Input error: Specify gene expression file. See --help for details')
                raise
            try:
                assert (self.gtf_file is not None)
            except AssertionError:
                print ('Input error: Specify GENCODE file. See --help for details')
                raise

            '''
            Check if files exist.
            '''
            for filepath in [self.vcf_file, self.oxf_file, self.fam_file, self.gx_file, self.gtf_file, self.gxcorr_file]:
                if (filepath is not None):
                    try:
                        assert os.path.isfile(filepath)
                    except AssertionError:
                        print ('File {:s} does not exist.'.format(filepath))
                        raise

            '''
            Check if output directory is writable / can be created
            '''
            outdir = os.path.dirname(os.path.realpath(self.outprefix))
            try:
                if not os.path.exists(outdir): os.makedirs(outdir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    print ('Unable to create output directory: {:s}'.format(outdir))
                    raise

            try:
                assert os.path.isdir(outdir) and os.access(outdir, os.W_OK | os.X_OK)
                #filepath = "{:s}.write_tester".format(self.outprefix)
                #filehandle = open( filepath, 'w' )
                #filehandle.close()
                #os.remove(filepath)
            except AssertionError:
                print('Unable to create files in {:s}'.format(outdir))
                raise



import unittest
import os
import subprocess
import pathlib

import tejaas.tester
  
class TestOutputFiles(unittest.TestCase):


    def __init__(self, *args, **kwargs):
        super(TestOutputFiles, self).__init__(*args, **kwargs)
        self.test_dir = pathlib.Path(__file__).parent.resolve()
        self.out_dir  = os.path.join(self.test_dir, "data")
        self.gold_dir = os.path.join(self.test_dir, "gold")
        self._run_bash_script(os.path.join(self.test_dir, "run_mwe.sh"))


    def test_rr_outfiles(self):
        for filename in ["result_rr.txt", "result_gene_snp_list.txt", "result_gene_snp_list_knn.txt"]:
            self._compare_with_gold_file(filename)
        return


    #def test_jpanull(self):
    #    self._compare_with_gold_file("result_jpanull.txt")
    #    return


    #def test_jpa_outfiles(self):
    #    for filename in ["result_fr_jpa.txt", "result_fr_gene_snp_list.txt"]:
    #        self._compare_with_gold_file(filename)
    #    return


    def _compare_with_gold_file(self, filename):
        out_filename = os.path.join(self.out_dir, filename)
        gold_filename = os.path.join(self.gold_dir, "gold_{:s}".format(filename))
        self._compare_files(out_filename, gold_filename)
        return
            

    def _compare_files(self, file1, file2):
        f1 = open(file1, "r")
        f2 = open(file2, "r")
        self.assertMultiLineEqual(f1.read(), f2.read())
        f1.close()
        f2.close()
        return


    def _run_bash_script(self, script_file):
        cmd     = ["bash", script_file, self.test_dir]
        with subprocess.Popen(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, bufsize=1, universal_newlines=True) as process:
            for line in process.stdout:
                print (line, end = "")
        retcode = process.returncode
        self.assertEqual(retcode, 0)
        return        

if __name__ == '__main__':
    tejaas.tester.main()



import numpy as np

def split_n(n, ncore):
    offset = 0
    endlist = list()
    startlist = list()
    nmax = int(n / ncore)
    for i in range(ncore-1):
        startlist.append(offset)
        endlist.append(offset + nmax)
        offset += nmax
    startlist.append(offset)
    endlist.append(n)
    return startlist, endlist

# convert this split as in split_genotype using split_n
def split_1darray(myarray, ncore):
    length = len(myarray)
    max_e = int(length / ncore)
    offset = 0
    data = [None for i in range(ncore)]
    for dest in range(ncore-1):
        start = offset
        end = offset + max_e
        data[dest] = myarray[start:end]
        offset += max_e 
    data[ncore-1] = myarray[offset:]
    return data


def split_genotype(geno, ncore):
    start, end = split_n(geno.shape[0], ncore)
    data = [geno[i:j, :] for i, j in zip(start, end)]
    return data, start
    

def split_genemasks(masks, nsnplist, offsetlist):
    res = None
    if masks is not None:
        res = [masks[offset : nsnp + offset] for offset, nsnp in zip(offsetlist, nsnplist)]
    return res


def split_maskcomp(masks, ncore):
    #nsnps = np.array([len(x.apply2) for x in masks])
    #start, end, _ = fuzzysplit(nsnps, ncore)
    nmask = len(masks)
    start, end = split_n(nmask, ncore)
    masksplit = [masks[i:j] for i,j in zip(start, end)]
    return masksplit


def fuzzysplit(narr, ncore):
    # A fuzzy splitting algorithm, to keep similar number of snps in each core.
    # Probably useless because SVD takes more time than calculation of each SNP.
    # Kept for legacy reason and backup.
    # Use this like:
    #   start, end, _ = fuzzysplit(nsnps, ncore)
    #   masksplit = [masks[i:j] for i,j in zip(start, end)]
    # It tries to distribute the remaining number of SNPs equally in all cores.
    # When confused, it decided a boundary based on the number of SNPs in the 2 adjacent masks.

    ncum = np.cumsum(narr)
    startlist = list()
    endlist = list()
    startlist.append(0)
    ntot = narr.shape[0]
    nrem = narr.shape[0]
    crem = ncore
    h = ncum[-1] / crem
    w = 0
    wold = 0
    for i, n in enumerate(narr):
        w += n
        nrem -= 1
        #print("Element {:d} [{:d}]. Sum = {:g}, Allowed = {:g}".format(i, n, w, h))
        #print("Remaining elements: {:d}. Remaining cores: {:d}".format(nrem, crem))
        if w >= h or nrem < crem:
            #if w >= h : print("Overflow add")
            #if nrem <= crem: print ("Constraint add")
            endix = 0
            _mw = 0
            if i > 0:
                excess = w - h
                deficit = h - wold
                endix = i
                _mw = 0
                #print ("Excess: {:g}, Deficit: {:g}".format(excess, deficit), i - 1, startlist[-1])
                if excess >= deficit and (i-1) >= startlist[-1]:
                    endix = i - 1
                    _mw = n
            w = _mw
            if nrem > 0:
                #print("Adding up to index {:d} in core {:d}. Carrying over: {:g}".format(endix, len(startlist), w))
                endlist.append(endix + 1)
                startlist.append(endix + 1)
            crem -= 1
            if crem > 0:
                h = (ncum[-1] - ncum[endix]) / crem
        wold = w
        #print("wold = {:g}".format(wold))
        
    #print("Adding up to index {:d} in core {:d}.".format(i, len(startlist)))
    endlist.append(narr.shape[0])
    arrsplit = [narr[i:j] for i,j in zip(startlist, endlist)]
    return startlist, endlist, arrsplit



import numpy as np
#from scipy.stats import chi2
from scipy.stats import norm
import logging
from tejaas.utils.logs import MyLogger

logger = MyLogger(__name__)

def perm_null(GT, GX, sigmabeta2, sigmax2):
    nsnps = GT.shape[0]
    nsamples = GT.shape[1]
    ngenes = GX.shape[0]
    Rscore = np.zeros(nsnps)
    pvals  = np.zeros(nsnps)
    muQ    = np.zeros(nsnps)
    sigmaQ = np.zeros(nsnps)

    logger.debug("Number of SNPs: {:d}".format(nsnps))
    logger.debug("Number of samples: {:d}".format(nsamples))
    logger.debug("Number of genes: {:d}".format(ngenes))

    Yt = GX.T # shape N x G
    U, S, Vt = np.linalg.svd(Yt, full_matrices=False)
    S2 = np.square(S)
    for i in range(nsnps):
        S2mod = S2 + sigmax2[i] / sigmabeta2[i]
        W = np.dot(U, np.dot(np.diag(S2 / S2mod), U.T)) #/ sigmax2[i]
        Rscore[i] = np.sum(np.square(np.dot(U.T, GT[i,:])) * S2 / S2mod)
        pvals[i], muQ[i], sigmaQ[i] = pvals_perm(GT[i, :].reshape(1, -1), Rscore[i], W)

    return pvals, Rscore, muQ, sigmaQ

def pvals_perm(GT, R, W):
    mu2, mu4 = moment_data(GT)
    N = GT.shape[1]
    q11 = np.sum(W)
    q2  = np.sum(np.diag(W))
    muQ = mu2 * (N * q2 - q11) / (N - 1)

    v31 = - mu4 / (N - 1)
    v22 = v31 + (N * mu2 * mu2 / (N - 1)) #(N*(mu2**2) - mu4)/(N-1)
    v211 = - (v31 + v22) / (N - 2)
    v1111 = - 3 * v211 / (N - 3)

    q31 = np.dot(np.diag(W),np.sum(W,axis = 1))
    q4 = np.sum(np.square(np.diag(W)))
    q22 = np.sum(np.square(W))
    q211 = np.sum(np.square(np.sum(W,axis = 1)))

    sigma2 = v1111*(q11**2 - 2*q2*q11 - 4*q211 + 8*q31 + 2*q22 + q2**2 - 6*q4) + 2*v211*(q2*q11 + 2*q211 - 6*q31 - 2*q22 - q2**2 + 6*q4) + v22*(q2**2 + 2*q22 - 3*q4) + 4*v31*(q31 - q4) + mu4*q4

    sigma2 = sigma2 - muQ**2
    #mscale = sigma2 / muQ / 2.0
    #df = muQ / mscale
    #Rscaled = R / mscale
    #p = 1 - chi2.cdf(Rscaled, df)
    sigmaQ = np.sqrt(sigma2)
    p = 1 - norm.cdf(R, loc=muQ, scale=sigmaQ)
    return p, muQ, sigmaQ

def moment_data(GT):   #GT ixN
    GT2 = np.square(GT)
    GT4 = np.square(GT2)
    mu2 = np.mean(GT2)
    mu4 = np.mean(GT4)
    return mu2, mu4


def maf_null(GT, GX, sigmabeta2, sigmax2, maf):
    nsnps = GT.shape[0]
    nsamples = GT.shape[1]
    ngenes = GX.shape[0]

    Yt = GX.T # shape N x G
    U, S, Vt = np.linalg.svd(Yt, full_matrices=False)
    S2 = np.square(S)
    S2mod = S2 + sigmax2[0] / sigmabeta2[0]
    W = np.dot(U, np.dot(np.diag(S2 / S2mod), U.T))
    Rscore = np.zeros(nsnps)
    for i in range(nsnps):
        Rscore[i] = np.sum(np.square(np.dot(U.T, GT[i,:])) * S2 / S2mod)
    pvals, muQ, sigmaQ = pvals_maf(S, Rscore, sigmabeta2, sigmax2, maf, W)
    return pvals, Rscore, muQ, sigmaQ


def pvals_maf(S, R, sigmabeta2, sigmax2, maf, W):
    nsnps = R.shape[0]
    S2 = np.square(S)
    S2mod = S2 + sigmax2[0] / sigmabeta2[0]
    mu = np.sum(S2 / S2mod)
    sigma2 = 2 * np.sum(np.square(S2 / S2mod))

    x4 = gt_fourth_moment(maf)
    corr_maf = (x4 - 3) * np.sum(np.diag(np.square(W)))
    sigma2 += corr_maf

    #mscale = sigma / mu / 2.0
    #df = mu / mscale
    #Rscaled = R / mscale
    #p = 1 - chi2.cdf(Rscaled, df)

    sigma = np.sqrt(sigma2)
    p = 1 - norm.cdf(R, loc=mu, scale=sigma)
    return p, np.repeat(mu, nsnps), sigma


def gt_fourth_moment(maf):
    f0 = np.square(1 - maf)
    f1 = 2.0 * maf * (1 - maf)
    f2 = np.square(maf)
    mu = 2 * maf
    sig = np.sqrt(f1)
    x4 = f0 *  (-mu/sig) ** 4 + f1 * ((1 - mu)/sig)**4 + f2 * ((2 - mu)/sig)**4
    return x4



import numpy as np
from collections import defaultdict
import gzip

from tejaas.utils.containers import GeneInfo, CisMask

'''
get_cismasklist
    Creates a list of genemasks for every SNP.
    A genemask is a list of gene_ids to be removed from analyses for each SNP.
compress_cismasklist
    Compress genemasks to CisMask containers.
    A CisMask container contains a set of gene_ids to be removed from a set of SNP_ids.
'''

def get_cismasklist(snpinfo, geneinfo, chrom, window=1e6):
    chr_genes_ix = [[] for ichrm in range(22)]
    chr_genes = [[] for ichrm in range(22)]
    if chrom is not None:
        chr_genes_ix[chrom - 1] = np.array([i for i, g in enumerate(geneinfo) if g.chrom == chrom])
        chr_genes[chrom - 1] = [geneinfo[ix] for ix in chr_genes_ix[chrom - 1]]
    else:
        for ichrm in range(22):
            chr_genes_ix[ichrm] = np.array([i for i, g in enumerate(geneinfo) if g.chrom == ichrm + 1])
            chr_genes[ichrm] = [geneinfo[ix] for ix in chr_genes_ix[ichrm]]
    genemasks = list()
    iprev = 0
    ichrmprev = 0
    for snp in snpinfo:
        pos = snp.bp_pos
        left = pos - window
        right = pos + window
        ichrm = chrom - 1 if chrom is not None else snp.chrom - 1
        iprev_started = False
        if ichrm != ichrmprev:
            iprev = 0
            ichrmprev = ichrm
        thismask = list()
        for i, g in enumerate(chr_genes[ichrm][iprev:]):
            gstart = g.start
            gend = g.end
            if gstart >= left and gstart <= right:
                # thismask.append(iprev + i)
                thismask.append(chr_genes_ix[ichrm][iprev + i])
                if not iprev_started:
                    new_start_iloc = iprev
                    iprev_started = True
            elif gend >= left and gend <= right:
                # thismask.append(iprev + i)
                thismask.append(chr_genes_ix[ichrm][iprev + i])
                if not iprev_started:
                    new_start_iloc = iprev
                    iprev_started = True
            if gstart > right:
                break
        if len(thismask) > 0:
            genemasks.append(np.array(thismask))
            iprev = new_start_iloc
        else:
            genemasks.append(np.array([]))
    return genemasks


def compress_cismasklist(genemasks):
    cismasks = list()
    appendmask = False
    endmask = False
    setprev = False
    snplist = list()
    for i, mask in enumerate(genemasks):
        if not setprev:
            prev_mask = mask
            setprev = True
        if np.all(np.array_equal(mask, prev_mask)):
            snplist.append(i)
        else:
            appendmask = True

        if i == len(genemasks) - 1: endmask = True # no more masks to process

        if appendmask:
            thismask = CisMask(rmv_id = prev_mask, apply2 = snplist)
            cismasks.append(thismask)
            snplist = list([i])
            prev_mask = mask
            if not endmask:
                appendmask = False

        if endmask:
            thismask = CisMask(rmv_id = mask, apply2 = snplist)
            cismasks.append(thismask)

    return cismasks

def read_crossmap(crossmapfile):
    gene_pairs = defaultdict(list)
    with gzip.open(crossmapfile, 'r') as instream:
        for line in instream:
            arr = line.rstrip().split()
            gene_pairs[arr[0].decode('utf-8')].append(arr[1].decode('utf-8'))
    return gene_pairs

def extend_cismask(genes, cismaskcomp, crossmapfile):
    cross_gene_dict = read_crossmap(crossmapfile)
    genes_ix_dict = dict(zip([x.ensembl_id for x in genes], np.arange(len(genes))))
    crossmap_cismaskcomp = list()
    for cismask in cismaskcomp:
        gene_list = list()  # list of crossmapped genes for the cismask
        for gene in [genes[i] for i in cismask.rmv_id]:
            gene_list += cross_gene_dict[gene.ensembl_id]
        uniq_gene_list = list(set(gene_list))
        ugene_dict = defaultdict(lambda: False) # significant speed-up by making a dict
        for g in uniq_gene_list:
            ugene_dict[g] = True
        cm_gene2rmv_ix = [genes_ix_dict[x.ensembl_id] for x in genes if ugene_dict[x.ensembl_id]]
        if len(cm_gene2rmv_ix) > 0:
            new_mask_genes = sorted(list(set(list(cismask.rmv_id) + cm_gene2rmv_ix)))
            new_cismask = cismask._replace(rmv_id = np.array(new_mask_genes))
            crossmap_cismaskcomp.append(new_cismask)
        else:
            crossmap_cismaskcomp.append(cismask)
    return crossmap_cismaskcomp
